/*
 * Copyright (c) 2004-2005 Sean Chittenden <sean@chittenden.org>
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Nexadesic: src/lib/libmemcache/src/memcache.c,v 1.78.2.7 2005/05/24 10:30:38 sean Exp $
 */

/* The crc32 functions and data was originally written by Spencer
 * Garrett <srg@quick.com> and was cleaned from the PostgreSQL source
 * tree via the files contrib/ltree/crc32.[ch].  No license was
 * included, therefore it is assumed that this code is public
 * domain.  Attribution still noted. */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <errno.h>
#include <sys/types.h>
#ifdef __linux
# ifndef __USE_POSIX
#  define __USE_POSIX
#warning "Working around busted-ass Linux header include problems: use FreeBSD instead"
#warning "http://www.FreeBSD.org/ - you won't regret it"
# endif
#endif
#include <sys/time.h>
#ifdef USE_MC_IOVEC
# define iovec iovec_old
# define iovec_cast struct iovec_old *
#else
# define iovec_cast struct iovec *
#endif
#include <sys/socket.h>
#include <sys/uio.h>
#ifdef USE_MC_IOVEC
# undef iovec
#endif
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>

/* Do some evilness to make sure that memcache(3) clients behave and
 * get warnings when they jack things up. */
#ifdef USE_MC_IOVEC
struct iovec {
  const void *iov_base;
  size_t iov_len;
};
#endif
#define __MEMCACHE__
# include "memcache.h"
#undef __MEMCACHE__

#ifdef MAX
# undef MAX
#endif
#define MAX(a,b) (((a)>(b))?(a):(b))

#ifdef MIN
# undef MIN
#endif
#define MIN(a,b) (((a)<(b))?(a):(b))


/* Makes it easier to read mcm_server_block() calls */
#define BLOCK	0
#define NBIO	1


/* Prototypes for static functions that are mcm_*() safe, but don't
 * require a memory context. */
static void			 mcm_reset_buf(struct memcache_server *ms);
static void			 mcm_server_init(const struct memcache_ctxt *ctxt, struct memcache_server *ms);

/* Prototypes for static functions that require a memory context */
static u_int32_t		 mcm_atomic_cmd(struct memcache_ctxt *ctxt, struct memcache *mc,
						const char *cmd, const size_t cmd_len,
						const char *key, const size_t key_len, const u_int32_t val);
static void			 mcm_err(const struct memcache_ctxt *ctxt, const u_int32_t flags,
					 const char *funcname, const u_int32_t lineno, const u_int32_t errcode,
					 const char *msg, const u_int32_t msglen, const u_int32_t errlvl);
static int32_t			 mcm_err_func(MCM_ERR_FUNC_SIG);
static void			 mcm_fetch_cmd(struct memcache_ctxt *ctxt, struct memcache *mc,
					       struct memcache_req *req, const char *cmd, const size_t cmd_len);
static char			*mcm_get_line(struct memcache_ctxt *ctxt, struct memcache *mc,
					      struct memcache_server *ms);
static void			 mcm_retrieve_data(struct memcache_ctxt *ctxt, struct memcache_req *req, struct memcache *mc, struct memcache_server *ms);
static void			 mcm_res_cb_free(struct memcache_req *req, struct memcache_res_cb *cb);
static struct memcache_res	*mcm_res_new(const struct memcache_ctxt *ctxt);
static void			 mcm_server_block(const struct memcache_ctxt *ctxt, struct memcache_server *ms, const int block);
static int			 mcm_server_connect(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms);
static struct memcache_server	*mcm_server_connect_next_avail(struct memcache_ctxt *ctxt, struct memcache *mc, const u_int32_t has);
static int			 mcm_server_readable(struct memcache_ctxt *ctxt, struct memcache_server *ms, struct timeval *tv);
static int			 mcm_server_resolve(struct memcache_ctxt *ctxt, struct memcache_server *ms);
static ssize_t			 mcm_server_send_cmd(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms, struct iovec *cmd, size_t cmdlen);
inline static ssize_t		 mcm_server_send_last_cmd(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms);
static struct memcache_server_stats	*mcm_server_stats_new(const struct memcache_ctxt *ctxt);
static int			 mcm_server_writable(struct memcache_ctxt *ctxt, struct memcache_server *ms, struct timeval *tv);
static int			 mcm_storage_cmd(struct memcache_ctxt *ctxt, struct memcache *mc,
						 const char *cmd, const size_t cmd_len,
						 const char *key, const size_t key_len,
						 const void *val, const size_t bytes,
						 const time_t expire, const u_int16_t flags);

/* Error handling convenience macros */
#define ERR_FLAG	0x1
#define WARN_FLAG	0x2
#define NO_ERRNO_FLAG	0x4
#define MCM_ERR(_code)			mcm_err(ctxt, ERR_FLAG, __FUNCTION__, __LINE__, _code, NULL, 0, 0)
#define MCM_ERRX(_code)			mcm_err(ctxt, ERR_FLAG|NO_ERRNO_FLAG, __FUNCTION__, __LINE__, _code, NULL, 0, 0)
#define MCM_ERR_MSG(_code, _msg)	mcm_err(ctxt, ERR_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), 0)
#define MCM_ERR_MSG_LVL(_code, _msg, _lvl)	mcm_err(ctxt, ERR_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), _lvl)
#define MCM_ERR_MSGX(_code, _msg)	mcm_err(ctxt, ERR_FLAG|NO_ERRNO_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), 0)
#define MCM_WARNX(_code, _msg)		mcm_err(ctxt, WARN_FLAG|NO_ERRNO_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), 0)
#define MCM_WARN_MSG(_code, _msg)	mcm_err(ctxt, WARN_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), 0)
#define MCM_WARN_MSGLEN(_code, _m, _l)	mcm_err(ctxt, WARN_FLAG, __FUNCTION__, __LINE__, _code, _m, _l, 0)
#define MCM_WARNX_MSG(_code, _msg)	mcm_err(ctxt, WARN_FLAG|NO_ERRNO_FLAG, __FUNCTION__, __LINE__, _code, _msg, strlen(_msg), 0)
#define MCM_WARNX_MSGLEN(_code, _m, _l)	mcm_err(ctxt, WARN_FLAG|NO_ERRNO_FLAG, __FUNCTION__, __LINE__, _code, _m, _l, 0)

#if defined(__STRICT_ANSI__) && !defined(__ISO_C_VISIBLE)
long long strtoll(const char *, char **, int);
#endif

/* This is kinda ugly, but, it saves on some warnings and a tad of
 * stack space across the library. Note, remember strlen(3) does not
 * include the trailing null character, but sizeoof() does, so when
 * computing the sizeof() commands, subtract one from its return. */
static const char	str_add_cmd[] = "add ";
static const size_t	str_add_len = MCM_CSTRLEN(str_add_cmd);
static const char	str_decr_cmd[] = "decr ";
static const size_t	str_decr_len = MCM_CSTRLEN(str_decr_cmd);
static const char	str_delete_cmd[] = "delete ";
static const size_t	str_delete_len = MCM_CSTRLEN(str_delete_cmd);
static const char	str_endl[] = "\r\n";
static const size_t	str_endl_len = MCM_CSTRLEN(str_endl);
static const char	str_get_cmd[] = "get ";
static const size_t	str_get_len = MCM_CSTRLEN(str_get_cmd);
static const char	str_incr_cmd[] = "incr ";
static const size_t	str_incr_len = MCM_CSTRLEN(str_incr_cmd);
#ifdef SEAN_HACKS
static const char	str_refresh_cmd[] = "refresh ";
static const size_t	str_refresh_len = MCM_CSTRLEN(str_refresh_cmd);
#endif
static const char	str_replace_cmd[] = "replace ";
static const size_t	str_replace_len = MCM_CSTRLEN(str_replace_cmd);
static const char	str_set_cmd[] = "set ";
static const size_t	str_set_len = MCM_CSTRLEN(str_set_cmd);
static const char	str_space[] = " ";
static const size_t	str_space_len = MCM_CSTRLEN(str_space);


/* #ifdef HAVE_KQUEUE */
/* /\* kqueue(2) is the preferred method for testing to see if a file */
/*  * descriptor has data on it. *\/ */
/* #else */
/* # ifdef HAVE_POLL */
/* /\* poll(2) is better than select(2), but still sucks*\/ */
/* # else */
/* #  ifdef HAVE_SELECT */
/* /\* If we have to, we'll use select(2). *\/ */
/* #  endif */
/* # endif */
/* #endif */

/* Set the default memory handling routines to be system defaults. */
static struct memcache_ctxt mcGlobalCtxt = {
  (mcFreeFunc)free,
  (mcMallocFunc)malloc,
  (mcMallocFunc)malloc,
  (mcReallocFunc)realloc,
  (mcErrFunc)mcm_err_func,
  0,
  NULL,
  0,
  0,
  0
};


int
mc_add(struct memcache *mc,
       const char *key, const size_t key_len,
       const void *val, const size_t bytes,
       const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(&mcGlobalCtxt, mc, str_add_cmd, str_add_len, key, key_len, val, bytes, expire, flags);
}


void *
mc_aget(struct memcache *mc, mc_const char *key, const size_t len) {
  return mcm_aget(&mcGlobalCtxt, mc, key, len);
}


#ifdef SEAN_HACKS
void *
mc_arefresh(struct memcache *mc, const char *key, const size_t len) {
  return mcm_arefresh(&mcGlobalCtxt, mc, key, len);
}
#endif


u_int32_t
mc_decr(struct memcache *mc, const char *key, const size_t key_len, const u_int32_t val) {
  return mcm_atomic_cmd(&mcGlobalCtxt, mc, str_decr_cmd, str_decr_len, key, key_len, val);
}


int
mc_delete(struct memcache *mc, const char *key, const size_t key_len, const time_t hold) {
  return mcm_delete(&mcGlobalCtxt, mc, key, key_len, hold);
}


int
mc_err_filter_add(const u_int32_t err_mask) {
  return mcm_err_filter_add(&mcGlobalCtxt, err_mask);
}


int
mc_err_filter_del(const u_int32_t err_mask) {
  return mcm_err_filter_del(&mcGlobalCtxt, err_mask);
}


u_int32_t
mc_err_filter_get(void) {
  return mcm_err_filter_get(&mcGlobalCtxt);
}


int
mc_err_filter_test(const u_int32_t err_lvl) {
  return mcm_err_filter_test(&mcGlobalCtxt, err_lvl);
}


void
mc_err_test(void) {
  mcm_err_test(&mcGlobalCtxt);
}


int
mc_flush(struct memcache *mc, struct memcache_server *ms) {
  return mcm_flush(&mcGlobalCtxt, mc, ms);
}


int
mc_flush_all(struct memcache *mc) {
  return mcm_flush_all(&mcGlobalCtxt, mc);
}


void
mc_free(struct memcache *mc) {
  mcm_free(&mcGlobalCtxt, mc);
}


void
mc_get(struct memcache *mc, struct memcache_req *req) {
  mcm_get(&mcGlobalCtxt, mc, req);
}


u_int32_t
mc_hash_key(const char *key, const size_t len) {
  return mcm_hash_key(&mcGlobalCtxt, key, len);
}


u_int32_t
mc_incr(struct memcache *mc, const char *key, const size_t key_len, const u_int32_t val) {
  return mcm_atomic_cmd(&mcGlobalCtxt, mc, str_incr_cmd, str_incr_len, key, key_len, val);
}


struct memcache *
mc_new(void) {
  return mcm_new(&mcGlobalCtxt);
}


#ifdef SEAN_HACKS
void
mc_refresh(struct memcache *mc, struct memcache_req *req) {
  mcm_fetch_cmd(&mcGlobalCtxt, mc, req, str_refresh_cmd, str_refresh_len);
}
#endif


const u_int32_t
mc_reldate(void) {
  return MEMCACHE_RELDATE;
}


int
mc_replace(struct memcache *mc,
	   const char *key, const size_t key_len,
	   const void *val, const size_t bytes,
	   const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(&mcGlobalCtxt, mc, str_replace_cmd, str_replace_len, key, key_len, val, bytes, expire, flags);
}


struct memcache_res *
mc_req_add(struct memcache_req *req, const char *key, const size_t len) {
  return mcm_req_add(&mcGlobalCtxt, req, key, len);
}


struct memcache_res *
mc_req_add_ref(struct memcache_req *req, mc_const char *key, const size_t len) {
  return mcm_req_add_ref(&mcGlobalCtxt, req, key, len);
}


void
mc_req_free(struct memcache_req *req) {
  mcm_req_free(&mcGlobalCtxt, req);
}


struct memcache_req *
mc_req_new(void) {
  return mcm_req_new(&mcGlobalCtxt);
}


int
mc_res_attempted(const struct memcache_res *res) {
  return mcm_res_attempted(&mcGlobalCtxt, res);
}


int
mc_res_found(const struct memcache_res *res) {
  return mcm_res_found(&mcGlobalCtxt, res);
}


void
mc_res_free(struct memcache_req *req, struct memcache_res *res) {
  mcm_res_free(&mcGlobalCtxt, req, res);
}


void
mc_res_free_on_delete(struct memcache_res *res, const int fod) {
  mcm_res_free_on_delete(&mcGlobalCtxt, res, fod);
}


int
mc_res_register_fetch_cb(struct memcache_req *req, struct memcache_res *res,
			 mcResCallback cb, void *misc) {
  return mcm_res_register_fetch_cb(&mcGlobalCtxt, req, res, cb, misc);
}


int
mc_server_activate(struct memcache *mc, struct memcache_server *ms) {
  return mcm_server_activate(&mcGlobalCtxt, mc, ms);
}


int
mc_server_activate_all(struct memcache *mc) {
  return mcm_server_activate_all(&mcGlobalCtxt, mc);
}


int
mc_server_add(struct memcache *mc, const char *hostname, const char *port) {
  return mcm_server_add2(&mcGlobalCtxt, mc, hostname, strlen(hostname), port, strlen(port));
}


int
mc_server_add2(struct memcache *mc, const char *hostname, const size_t hostname_len,
	       const char *port, const size_t port_len) {
  return mcm_server_add2(&mcGlobalCtxt, mc, hostname, hostname_len, port, port_len);
}


int
mc_server_add3(struct memcache *mc, struct memcache_server *ms) {
  return mcm_server_add3(&mcGlobalCtxt, mc, ms);
}


int
mc_server_add4(struct memcache *mc, mc_const char *hostport) {
  return mcm_server_add5(&mcGlobalCtxt, mc, hostport, strlen(hostport));
}


int
mc_server_add5(struct memcache *mc, mc_const char *hostport, const size_t hostlen) {
  return mcm_server_add5(&mcGlobalCtxt, mc, hostport, hostlen);
}


void
mc_server_deactivate(struct memcache *mc, struct memcache_server *ms) {
  mcm_server_deactivate(&mcGlobalCtxt, mc, ms);
}


void
mc_server_disconnect(struct memcache_server *ms) {
  mcm_server_disconnect(&mcGlobalCtxt, ms);
}


void
mc_server_disconnect_all(const struct memcache *mc) {
  mcm_server_disconnect_all(&mcGlobalCtxt, mc);
}


struct memcache_server *
mc_server_find(struct memcache *mc, const u_int32_t hash) {
  return mcm_server_find(&mcGlobalCtxt, mc, hash);
}


void
mc_server_free(struct memcache_server *ms) {
  mcm_server_free(&mcGlobalCtxt, ms);
}


struct memcache_server *
mc_server_new(void) {
  return mcm_server_new(&mcGlobalCtxt);
}


struct memcache_server_stats *
mc_server_stats(struct memcache *mc, struct memcache_server *ms) {
  return mcm_server_stats(&mcGlobalCtxt, mc, ms);
}


void
mc_server_stats_free(struct memcache_server_stats *s) {
  mcm_server_stats_free(&mcGlobalCtxt, s);
}


int
mc_set(struct memcache *mc,
       const char *key, const size_t key_len,
       const void *val, const size_t bytes,
       const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(&mcGlobalCtxt, mc, str_set_cmd, str_set_len, key, key_len, val, bytes, expire, flags);
}


struct memcache_server_stats *
mc_stats(struct memcache *mc) {
  return mcm_stats(&mcGlobalCtxt, mc);
}


char *
mc_strdup(const char *str) {
  return mcm_strndup(&mcGlobalCtxt, str, strlen(str));
}


char *
mc_strnchr(mc_const char *str, const int c, const size_t len) {
  return mcm_strnchr(&mcGlobalCtxt, str, c, len);
}


char *
mc_strndup(const char *str, const size_t len) {
  return mcm_strndup(&mcGlobalCtxt, str, len);
}


void
mc_timeout(struct memcache *mc, const int sec, const int msec) {
  mcm_timeout(&mcGlobalCtxt, mc, sec, msec);
}


const u_int32_t
mc_vernum(void) {
  return MEMCACHE_VERNUM;
}


const char *
mc_version(void) {
  return MEMCACHE_VER;
}
/* END OF THE SINGLE MEMORY CONTEXT API CALLS (ie: mc_*()) */


/* BEGIN MEMORY CONTEXT API (ie: mcm_*())  */
int
mcm_add(struct memcache_ctxt *ctxt, struct memcache *mc,
	const char *key, const size_t key_len,
	const void *val, const size_t bytes,
	const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(ctxt, mc, str_add_cmd, str_add_len, key, key_len, val, bytes, expire, flags);
}


/* Issues a "get" command to the memcache server that should contain
 * the key.  The result is mcMalloc(3)'ed and it is assumed that the
 * caller is required to mcFree(3) the memory. */
void *
mcm_aget(struct memcache_ctxt *ctxt, struct memcache *mc, mc_const char *key, const size_t len) {
  struct memcache_req *req;
  struct memcache_res *res;
  void *ret;

  req = mcm_req_new(ctxt);
  res = mcm_req_add_ref(ctxt, req, key, len);
  mcm_res_free_on_delete(ctxt, res, 0);
  mcm_get(ctxt, mc, req);
  ret = res->val;
  mcm_req_free(ctxt, req);
  return ret;
}


#ifdef SEAN_HACKS
/* Issues a "refresh" command to the memcache server that should
 * contain the key.  The result is mcMalloc(3)'ed and it is assumed
 * that the caller is required to mcFree(3) the memory. */
void *
mcm_arefresh(struct memcache_ctxt *ctxt, struct memcache *mc, const char *key, const size_t len) {
  struct memcache_req *req;
  struct memcache_res *res;
  void *ret;

  req = mcm_req_new(ctxt);
  res = mcm_req_add_ref(ctxt, req, key, len);
  mcm_res_free_on_delete(ctxt, res, 0);
  mcm_refresh(ctxt, mc, req);
  ret = res->val;
  mcm_req_free(ctxt, req);
  return ret;
}
#endif


static u_int32_t
mcm_atomic_cmd(struct memcache_ctxt *ctxt, struct memcache *mc,
	       const char *cmd, const size_t cmd_len,
	      const char *key, const size_t key_len, const u_int32_t val) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cp, *cur;
  size_t i;
  struct iovec av[5];
  u_int32_t ret;

  /* Reset errnum on re-entry into memcache(3). */
  ctxt->errnum = 0;

  /* If we have only one server, don't bother actually hashing. */
  if (mc->num_live_servers == 1)
    hash = 0;
  else
    hash = mcm_hash_key(ctxt, key, key_len);

  ms = mcm_server_connect_next_avail(ctxt, mc, hash);
  if (ms == NULL) {
    MCM_ERRX(MCM_ERR_MC_VALID_SERVER);
    return 0;
  }

  mcm_reset_buf(ms);
  ms->woff = 0;

  av[0].iov_base = cmd;
  av[0].iov_len = cmd_len;
  av[1].iov_base = key;
  av[1].iov_len = key_len;
  av[2].iov_base = str_space;
  av[2].iov_len = str_space_len;


  /* 10 == (2 ** 32).to_s.length */
  if (ms->size - ms->woff < 10) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 10));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return 0;
    } else {
      ms->size += MAX(ms->size, 10);
      ms->buf = cp;
    }
  }

  /* Convert the value to a string */
  i = (size_t)snprintf(&ms->buf[ms->woff], ms->size - ms->woff, "%u", val);
  if (i < 1) {
    MCM_ERR(MCM_ERR_LIB_SNPRINTF);
    return 0;
  }

  av[3].iov_base = &ms->buf[ms->woff];
  av[3].iov_len = i;
  av[4].iov_base = str_endl;
  av[4].iov_len = str_endl_len;

  ms->woff += ++i;

  mcm_server_send_cmd(ctxt, mc, ms, av, 5);

  mcm_reset_buf(ms);
  cur = mcm_get_line(ctxt, mc, ms);
  if (cur != NULL && memcmp(cur, "NOT_FOUND", MCM_CSTRLEN("NOT_FOUND")) == 0) {
    ctxt->errnum = ENOENT;
    return 0;
  } else if (cur == NULL) {
    return 0;
  }


  /* 10 == (2 ** 32).to_s.length */
  if (ms->size - ms->woff < 10) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 10));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return 0;
    } else {
      ms->size += MAX(ms->size, 10);
      cur = ms->buf = cp;
    }
  }

  /* Try converting the value to an integer. If it succeeds, we've got
   * a winner. */
  ret = (u_int32_t)strtol(cur, &cp, 10);
  if (ret == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
    MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "strtol(3) failed");
    return 0;
  }
  ms->soff += cp - cur;

#ifdef DEBUG_MC_PROTO_ASSERT
  if (*cp != '\r') {
    MCM_ERRX(MCM_ERR_PROTO);
    return 0;
  }
#endif

  return ret;
}


u_int32_t
mcm_decr(struct memcache_ctxt *ctxt, struct memcache *mc, const char *key, const size_t key_len, const u_int32_t val) {
  return mcm_atomic_cmd(ctxt, mc, str_decr_cmd, str_decr_len, key, key_len, val);
}


int
mcm_delete(struct memcache_ctxt *ctxt, struct memcache *mc,
	   const char *key, const size_t key_len, const time_t hold) {
  struct memcache_server *ms;
  u_int32_t hash;
  char *cp;
  size_t i;
  struct iovec dv[5];

  /* Reset ctxt->errnum upon entry into memcache(3). */
  ctxt->errnum = 0;

  /* If we have only one server, don't bother actually hashing. */
  if (mc->num_live_servers == 1)
    hash = 0;
  else
    hash = mcm_hash_key(ctxt, key, key_len);

  ms = mcm_server_connect_next_avail(ctxt, mc, hash);
  if (ms == NULL)
    return -1;

  mcm_reset_buf(ms);
  ms->woff = 0;

  dv[0].iov_base = str_delete_cmd;
  dv[0].iov_len = str_delete_len;
  dv[1].iov_base = key;
  dv[1].iov_len = key_len;
  dv[2].iov_base = str_space;
  dv[2].iov_len = str_space_len;


  /* 10 == (2 ** 32).to_s.length */
  if (ms->size - ms->woff < 10) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 10));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return -2;
    } else {
      ms->size += MAX(ms->size, 10);
      ms->buf = cp;
    }
  }

  /* Convert expiration time to a string */
  i = (size_t)snprintf(&ms->buf[ms->woff], ms->size - ms->woff, "%u", (u_int32_t)hold);
  if (i < 1) {
    MCM_ERR(MCM_ERR_LIB_SNPRINTF);
    return -4;
  }

  dv[3].iov_base = &ms->buf[ms->woff];	/* Note where our flags string is located */
  dv[3].iov_len = i;
  dv[4].iov_base = str_endl;
  dv[4].iov_len = str_endl_len;

  ms->woff += ++i;

  if (mcm_server_send_cmd(ctxt, mc, ms, dv, 5) < 0)
    return -3;

  mcm_reset_buf(ms);
  cp = mcm_get_line(ctxt, mc, ms);
  if (cp != NULL && memcmp(cp, "DELETED", MCM_CSTRLEN("DELETED")) == 0) {
    return 0;
  } else if (cp != NULL && memcmp(cp, "NOT_FOUND", MCM_CSTRLEN("NOT_FOUND")) == 0) {
    return 1;
  } else {
    MCM_ERRX(MCM_ERR_PROTO);
    return -5;
  }
}


static void
mcm_err(const struct memcache_ctxt *ctxt, const u_int32_t flags, const char *funcname, const u_int32_t lineno,
	const u_int32_t errcode, const char *msg, const u_int32_t msglen, const u_int32_t errlvl) {
  struct memcache_err_ctxt ectxt;

  bzero(&ectxt, sizeof(struct memcache_err_ctxt));

  ectxt.funcname = funcname;
  ectxt.lineno = lineno;
  ectxt.errnum = ((flags & NO_ERRNO_FLAG) ? 0 : errno);
  ectxt.errcode = errcode;
  ectxt.errmsg = msg;
  ectxt.errlen = msglen;

  /* Collect all error handling into one place and dispatch handlers
   * from here. */
  switch(errcode) {
  case MCM_ERR_NONE:
    ectxt.errstr = "no error";
    ectxt.severity = MCM_ERR_LVL_NONE;
    ectxt.sysexit = EX_OK;
    break;
  case MCM_ERR_ASSERT:
    ectxt.errstr = "internal memcache(3) assertion";
    ectxt.severity = MCM_ERR_LVL_FATAL;
    ectxt.sysexit = EX_SOFTWARE;
    break;
  case MCM_ERR_LIB_SNPRINTF:
    ectxt.errstr = "snprintf(3) failed to convert the value to a string";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_DATAERR;
    break;
  case MCM_ERR_LIB_STRTOL:
    ectxt.errstr = "strtol(3) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_DATAERR;
    break;
  case MCM_ERR_LIB_STRTOLL:
    ectxt.errstr = "strtoll(3) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_DATAERR;
    break;
  case MCM_ERR_MC_RECONN:
    ectxt.errstr = "connection re-established with server";
    ectxt.severity = MCM_ERR_LVL_INFO;
    ectxt.sysexit = EX_OK;
    break;
  case MCM_ERR_MC_SEND_CMD:
    ectxt.errstr = "failed to send command to the memcache server";
    ectxt.severity = MCM_ERR_LVL_NOTICE;
    ectxt.sysexit = EX_IOERR;
    break;
  case MCM_ERR_MC_SERV_LIST:
    ectxt.errstr = "no available servers in server list";
    ectxt.severity = MCM_ERR_LVL_WARN;
    ectxt.sysexit = EX_DATAERR;
    break;
  case MCM_ERR_MC_STORE:
    ectxt.errstr = "unable to store value";
    ectxt.severity = MCM_ERR_LVL_NOTICE;
    ectxt.sysexit = EX_CANTCREAT;
    break;
  case MCM_ERR_MC_VALID_SERVER:
    ectxt.errstr = "unable to find a server to connect to";
    ectxt.severity = MCM_ERR_LVL_NOTICE;
    ectxt.sysexit = EX_UNAVAILABLE;
    break;
  case MCM_ERR_MEM_MALLOC:
    ectxt.errstr = "mcMalloc(3) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_MEM_REALLOC:
    ectxt.errstr = "mcRealloc(3) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_NET_CONNECT:
    ectxt.errstr = "unable to connect to a server";
    ectxt.severity = MCM_ERR_LVL_NOTICE;
    ectxt.sysexit = EX_TEMPFAIL;
    break;
  case MCM_ERR_NET_HOST:
    ectxt.errstr = "unable to lookup/resolve host";
    ectxt.severity = MCM_ERR_LVL_WARN;
    ectxt.sysexit = EX_NOHOST;
    break;
  case MCM_ERR_PROTO:
    ectxt.errstr = "memcache(4) protocol error";
    ectxt.severity = MCM_ERR_LVL_FATAL;
    ectxt.sysexit = EX_PROTOCOL;
    break;
  case MCM_ERR_SYS_CLOSE:
    ectxt.errstr = "close(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_CONNECT:
    ectxt.errstr = "connect(2) failed";
    ectxt.severity = MCM_ERR_LVL_NOTICE;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_FCNTL:
    ectxt.errstr = "unable to get or set file descriptor status";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_READ:
    ectxt.errstr = "read(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_SELECT:
    ectxt.errstr = "select(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_SETSOCKOPT:
    ectxt.errstr = "setsockopt(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_SOCKET:
    ectxt.errstr = "socket(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
    break;
  case MCM_ERR_SYS_WRITEV:
    ectxt.errstr = "writev(2) failed";
    ectxt.severity = MCM_ERR_LVL_ERR;
    ectxt.sysexit = EX_OSERR;
  case MCM_ERR_TEST:
    ectxt.errstr = "internal memcache(3) test message";
    ectxt.severity = MCM_ERR_LVL_WARN;
    ectxt.sysexit = EX_OK;
    break;
  case MCM_ERR_TIMEOUT:
    ectxt.errstr = "timeout";
    ectxt.severity = MCM_ERR_LVL_WARN;
    ectxt.sysexit = EX_UNAVAILABLE;
    break;
  case MCM_ERR_TRACE:
    ectxt.errstr = "memcache(3) trace";
    ectxt.severity = MCM_ERR_LVL_INFO;
    ectxt.sysexit = EX_OK;
    break;
  case MCM_ERR_UNKNOWN_STAT:
    ectxt.errstr = "unknown stat variable";
    ectxt.severity = MCM_ERR_LVL_WARN;
    ectxt.sysexit = EX_PROTOCOL;
    break;
  default:
    ectxt.errstr = "unknown error code";
    ectxt.severity = MCM_ERR_LVL_FATAL;
    ectxt.sysexit = EX_SOFTWARE;
  }

  /* If we were passed in an error level, override the default
   * severity. */
  if (errlvl != 0)
    ectxt.severity = errlvl;

  /* Apply the error filter and ignore errors from levels that are
   * ignored. */
  if ((ctxt->MCM_ERR_MASK & ectxt.severity) != 0)
    return;

  /* Determine whether or not we continue running depending on the severity */
  switch (ectxt.severity) {
  case MCM_ERR_LVL_INFO:
  case MCM_ERR_LVL_NOTICE:
  case MCM_ERR_LVL_WARN:
    ectxt.cont = 'y';
    break;
  case MCM_ERR_LVL_ERR:
    ectxt.cont = 'n';
    break;
  case MCM_ERR_LVL_FATAL:
  default:
    ectxt.cont = 'a';
  }

  /* Call the user's handler.  Disregard the return value for now, but
   * have it there for future use.  *shrug* */
  if (ctxt->mcErr != NULL)
    (void)ctxt->mcErr(ctxt, &ectxt);

  /* There are a few error codes that require special cases for */
  switch (errcode) {
  case MCM_ERR_MC_SERV_LIST:
    if (ectxt.cont == 'n')
      ectxt.cont = 'y';
    break;
  }

  switch (ectxt.cont) {
  case 'y':
    /* Yes: do nothing */
    break;
  case 'n':
    /* No: exit with an error code */
    exit(ectxt.sysexit);
  case 'a':
    /* Abort: do just that, abort(3) */
  default:
    abort();
  }
}


int
mcm_err_filter_add(struct memcache_ctxt *ctxt, const u_int32_t err_mask) {
  if ((ctxt->MCM_ERR_MASK & err_mask) == ctxt->MCM_ERR_MASK)
    return 0;

  ctxt->MCM_ERR_MASK &= err_mask;
  return 1;
}


int
mcm_err_filter_del(struct memcache_ctxt *ctxt, const u_int32_t err_mask) {
  if ((ctxt->MCM_ERR_MASK & err_mask) == ctxt->MCM_ERR_MASK)
    return 0;

  ctxt->MCM_ERR_MASK &= ~err_mask;
  return 1;
}


u_int32_t
mcm_err_filter_get(const struct memcache_ctxt *ctxt) {
  return ctxt->MCM_ERR_MASK;
}


int
mcm_err_filter_test(const struct memcache_ctxt *ctxt, const u_int32_t err_lvl) {
  return(((ctxt->MCM_ERR_MASK & err_lvl) != 0) ? 1 : 0);
}


static int32_t
mcm_err_func(MCM_ERR_FUNC_ARGS) {
  const struct memcache_ctxt *ctxt;
  struct memcache_err_ctxt *ectxt;
  const char *errno_str, *severity;
  struct timeval tv;

  MCM_ERR_INIT_CTXT(ctxt, ectxt);

  if (ectxt->errnum != 0)
    errno_str = strerror(ectxt->errnum);
  else
    errno_str = NULL;

  switch (ectxt->severity) {
  case MCM_ERR_LVL_INFO:
    severity = "INFO";
    break;
  case MCM_ERR_LVL_NOTICE:
    severity = "NOTICE";
    break;
  case MCM_ERR_LVL_WARN:
    severity = "WARN";
    break;
  case MCM_ERR_LVL_ERR:
    severity = "ERROR";
    break;
  case MCM_ERR_LVL_FATAL:
    severity = "FATAL";
    break;
  default:
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      tml = asprintf(&tm, "Unknown error severity: %d", ectxt->severity);
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while(0);
#endif
    severity = "UNKNOWN";
  }

  /*
   * Quick explaination of the various bits of text:
   *
   * ectxt->errmsg - per error message passed along via one of the MCM_*_MSG() macros (optional)
   * ectxt->errstr - memcache(3) error string (optional, though almost always set)
   * errno_str - errno error string (optional)
   */

  if (gettimeofday(&tv, NULL) != 0) {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
  }

  if (ectxt->errmsg != NULL && errno_str != NULL && ectxt->errmsg != NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s: %s: %.*s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, ectxt->errstr, errno_str, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg == NULL && errno_str != NULL && ectxt->errmsg != NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s: %.*s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, errno_str, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg != NULL && errno_str == NULL && ectxt->errmsg != NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s: %.*s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, ectxt->errstr, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg != NULL && errno_str != NULL && ectxt->errmsg == NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s: %s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, errno_str, ectxt->errstr);
  else if (ectxt->errmsg == NULL && errno_str == NULL && ectxt->errmsg != NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %.*s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, (int)ectxt->errlen, ectxt->errmsg);
  else if (ectxt->errmsg == NULL && errno_str != NULL && ectxt->errmsg == NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, errno_str);
  else if (ectxt->errmsg != NULL && errno_str == NULL && ectxt->errmsg == NULL)
    fprintf(stderr, "[%s@%d.%06d] %s():%u: %s\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno, ectxt->errmsg);
  else
    fprintf(stderr, "[%s@%d.%06d] %s():%u\n", severity, (int)tv.tv_sec, (int)tv.tv_usec, ectxt->funcname, ectxt->lineno);

  return 0;
}


void
mcm_err_test(const struct memcache_ctxt *ctxt) {
  MCM_WARNX(MCM_ERR_TEST, "per-error message specific to this line of code");
}


static void
mcm_fetch_cmd(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_req *req,
	      const char *cmd, const size_t cmd_len) {
  struct memcache_res *res;
  struct memcache_res_cb *cb;
  struct memcache_server *ms;
  struct iovec *rv;
  u_int32_t i, num_vec;

  if (req->num_keys == 0)
    return;

  /* Need two extra iovec structs, one for the command, one for the
   * trailing \r\n. */
  num_vec = 2 * req->num_keys + 1;
  rv = (struct iovec *)ctxt->mcMalloc(sizeof(struct iovec) * num_vec);
  rv[0].iov_base = cmd;
  rv[0].iov_len = cmd_len;
  for (i = 1, res = req->query.tqh_first; res != NULL; res = res->entries.tqe_next, i++) {
    /* If we have only one live server, don't waste CPU actually
       hashing. */
    if (mc->num_live_servers == 1)
      res->hash = 0;
    else if (res->hash == 0)
      res->hash = mcm_hash_key(ctxt, res->key, res->len);

    rv[i].iov_base = res->key;
    rv[i].iov_len = res->len;

    if (res->entries.tqe_next != NULL) {
      rv[++i].iov_base = str_space;
      rv[i].iov_len = 1;
    }

    /* While we're looping, might as well see if we should be auto
     * deleting any of these keys. */
    if ((res->_flags & (MCM_RES_FREE_ON_DELETE | MCM_RES_NO_FREE_ON_DELETE)) ==
	(MCM_RES_FREE_ON_DELETE | MCM_RES_FREE_ON_DELETE))
      mcm_res_free_on_delete(ctxt, res, (res->size > 0 ? 0 : 1));
  }
  rv[i].iov_base = str_endl;
  rv[i].iov_len = str_endl_len;

  /* XXX Yuk.  Right now I'm grabbing the hash from the first key and
   * using that, but that's dangerous in the event that the second key
   * in the request could be stored on a different server.  memcache
   * servers need the ability to query other servers on the users
   * behalf, IMHO. */
  ms = mcm_server_connect_next_avail(ctxt, mc, req->query.tqh_first->hash);
  if (ms == NULL) {
    ctxt->mcFree(rv);
    return;
  }

  if (mcm_server_send_cmd(ctxt, mc, ms, rv, num_vec) < 0) {
    ctxt->mcFree(rv);
    return;
  }

  /* Reset our buffer */
  mcm_reset_buf(ms);

  /* Grab data from the server */
  mcm_retrieve_data(ctxt, req, mc, ms);

  for (res = req->query.tqh_first; res != NULL; res = res->entries.tqe_next) {
    /* Set the attempted bit.  This is rather late in the process, but
     * setting the bit here allows for server failures to not count as
     * an attempt */
    res->_flags |= MCM_RES_ATTEMPTED;
  }

  /* Now that we've finished the IO, fire off any callbacks that are
   * registered. */
  for (cb = req->cb.tqh_first; cb != NULL; cb = cb->entries.tqe_next) {
    (*cb->cb)(cb->ctxt, cb->res, cb->misc);
  }

  ctxt->mcFree(rv);
  return;
}


int
mcm_flush(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  char *cur;
  struct iovec fv[1];

  if (mcm_server_connect(ctxt, mc, ms) == -1)
    return -1;

  fv[0].iov_base = "flush_all\r\n";
  fv[0].iov_len = MCM_CSTRLEN("flush_all\r\n");

  if (mcm_server_send_cmd(ctxt, mc, ms, fv, 1) < 0)
    return -2;

  mcm_reset_buf(ms);
  cur = mcm_get_line(ctxt, mc, ms);
  if (cur != NULL && memcmp(cur, "OK", MCM_CSTRLEN("OK")) == 0) {
    return 0;
  } else {
    MCM_ERRX(MCM_ERR_PROTO);
    return -3;
  }
}


int
mcm_flush_all(struct memcache_ctxt *ctxt, struct memcache *mc) {
  struct memcache_server *ms;
  int ret = 0,
    tret;

  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next) {
    tret = mcm_flush(ctxt, mc, ms);

    /* Return the error code of the first non-zero value if there is
     * one.  Not sure if this is correct, but I don't have a better
     * idea right now. XXX */
    if (tret != 0 && ret == 0)
      ret = tret;
  }

  return ret;
}


void
mcm_free(const struct memcache_ctxt *ctxt, struct memcache *mc) {
  struct memcache_server *ms, *tms;

  if (mc == NULL)
    return;

  tms = mc->server_list.tqh_first;
  while(tms != NULL) {
    ms = tms;
    tms = ms->entries.tqe_next;

    mcm_server_free(ctxt, ms);
  }

  if (mc->live_servers != NULL) {
    ctxt->mcFree(mc->live_servers);
  }

  ctxt->mcFree(mc);
}


void
mcm_get(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_req *req) {
  /* At this point, we can safely assume we're performing a multi-get.
   * Pre-calculate what keys map up with which servers in order to
   * group requests where possible. Assume the worst/best case
   * scenario in that all keys will go to one server.*/
  struct memcache_req	**psq, *tsq;	/* Per-Server-reQuest */
  struct memcache_res	*psr, *trs;	/* Per-Server-Response */
  const u_int16_t	 num_servers = mc->num_live_servers;
  u_int16_t		 i;

  /* Reset ctxt->errnum upon entry into memcache(3), even though
   * mcm_get() and its call graph doesn't make use of ctxt->errnum. */
  ctxt->errnum = 0;

  /* Perform some trickery and call mcm_fetch_cmd() once for every
   * server that needs to be queried.  Short-circuit execution where
   * possible by invoking mcm_fetch_cmd() immediately if there is only
   * one key being queried. */
  switch (req->num_keys) {
  case 0:
    return;
  case 1:
    mcm_fetch_cmd(ctxt, mc, req, str_get_cmd, str_get_len);
    return;
  }

  /* If we're a multi-get but there is only one server, don't bother
   * with splitting out the keys to their appropriate server. */
  switch(mc->num_live_servers) {
  case 0:
    return;
  case 1:
    mcm_fetch_cmd(ctxt, mc, req, str_get_cmd, str_get_len);
    return;
  }

  /* Create an array of pointers to the per-server request objects.
   * Allocate one extra request object as a terminator of the pointer
   * array. */
  psq = (struct memcache_req**)ctxt->mcMalloc(sizeof(struct memcache_req*) * (num_servers + 1));
  if (psq == NULL) {
    MCM_ERR_MSG(MCM_ERR_ASSERT, "memory was not allocated for psq");
    return;
  }
  bzero(psq, sizeof(struct memcache_req*) * (num_servers + 1));

  /* Create a request object for each server. */
  for (i = 0; i < num_servers; i++) {
    psq[i] = mcm_req_new(ctxt);
  }

  /* Make a first pass through the keys to determine which server they
   * belong to. */
  for (trs = req->query.tqh_first; trs != NULL; trs = trs->entries.tqe_next) {
    psr = mcm_res_new(ctxt);

    /* Shallow copy of trs into psr. */
    psr->key = trs->key;
    psr->len = trs->len;
    psr->hash = trs->hash;
    psr->val = trs->val;
    psr->bytes = trs->bytes;
    psr->size = trs->size;
    psr->flags = trs->flags;

    /* No flags for shadow structure: we don't want the key or value
     * to be reaped when we cleanup. */
    psr->_flags = 0;

    mcm_res_free_on_delete(ctxt, psr, 0);

    if (psr->hash == 0 && mc->num_live_servers > 1) {
      psr->hash = trs->hash = mcm_hash_key(ctxt, psr->key, psr->len);
    }

    /* Store a pointer to the original object. */
    psr->misc = trs;

    /* Append onto the correct request chain for a server. */
    tsq = psq[psr->hash % num_servers];
    TAILQ_INSERT_TAIL(&tsq->query, psr, entries);
    tsq->num_keys++;
  }

  /* Make a second pass through the list of requests and execute the
   * fetch command where appropriate. */
  for (i = 0; i < num_servers; i++) {
    if (psq[i]->num_keys == 0)
      continue;

    mcm_fetch_cmd(ctxt, mc, psq[i], str_get_cmd, str_get_len);

    /* Copy the important bits back. */
    for (psr = psq[i]->query.tqh_first; psr != NULL; psr = psr->entries.tqe_next) {
      trs = (struct memcache_res*)psr->misc;
      trs->val = psr->val;
      trs->bytes = psr->bytes;
      trs->size = psr->size;
      trs->flags = psr->flags;
      trs->_flags |= psr->_flags;
    }
  }

  /* Cleanup */
  for (i = 0; i < num_servers; i++) {
    mcm_req_free(ctxt, psq[i]);
  }
  ctxt->mcFree(psq);
}


static char *
mcm_get_line(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  ssize_t rb;
  char *cp, *cur;

  if (ms->woff == -1) {
    /* Can't have a negative offset: -1 is only used to suggest that
     * we need to read(2) data. */
    ms->woff = 0;

    try_read:
    mcm_server_block(ctxt, ms, NBIO);
    if (mcm_server_readable(ctxt, ms, &ms->tv) < 0) {
      mcm_server_deactivate(ctxt, mc, ms);
      MCM_ERR_MSG(MCM_ERR_SYS_SELECT, "select returned bogus value");
      return NULL;
    }

    rb = read(ms->fd, &ms->buf[ms->woff], ms->size - ms->woff);
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      tml = asprintf(&tm, "%d = read(2)'s buf: \"%.*s\", errno = %d", (int)rb, (int)MIN((ssize_t)(ms->size - ms->woff), rb), &ms->buf[ms->woff], errno);
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif
    switch(rb) {
    case -1:
      /* We're in non-blocking mode, don't abort because of EAGAIN or
       * EINTR */
      switch (errno) {
      case EAGAIN:
      case EINTR:
	goto try_read;
      case ECONNRESET:
      case EINVAL:
	/* The server was restarted.  We shouldn't deactivate the
	 * server, but we should try and reconnect to it and roll from
	 * there. */
	MCM_ERR_MSG_LVL(MCM_ERR_SYS_READ, strerror(errno), MCM_ERR_LVL_INFO);

	mcm_server_disconnect(ctxt, ms);

	/* Reconnect to the same server.  If this fails, get the next
	 * available server. */
	if (mcm_server_connect(ctxt, mc, ms) == -1) {
	  mcm_server_deactivate(ctxt, mc, ms);
	  ms = mcm_server_connect_next_avail(ctxt, mc, ms->_last_hash);

	  if (ms == NULL)
	    return NULL;
	} else {
	  MCM_ERRX(MCM_ERR_MC_RECONN);
	}

	/* Need to send the last request again that way we have
	 * something to read(2). */
	mcm_server_send_last_cmd(ctxt, mc, ms);

	ms->woff = ms->startoff = ms->soff = 0;
	goto try_read;
      case EBADF:
      case EFAULT:
	/* This shouldn't happen and if it does, we're pooched: better
	 * dump. */
	MCM_ERR_MSG_LVL(MCM_ERR_SYS_READ, strerror(errno), MCM_ERR_LVL_FATAL);
	break;

      default:
	/* We shouldn't ever get here. */
	MCM_ERR_MSG(MCM_ERR_ASSERT, strerror(errno));
	return NULL;
      }
    case 0:
      /* Seems like an error to me if the server closes its connection
       * here.  deactivate the server instead of just closing it. */
      mcm_server_deactivate(ctxt, mc, ms);
      MCM_ERR_MSG(MCM_ERR_SYS_READ, "server unexpectedly closed connection");
      return NULL;
    default:
      ms->woff += rb;
    }
    /* If we got the same amount of data as we said was max, then we
     * probably need to read(2) more data.  As such we will double
     * buffer size right now.  Worst case scenario we double it
     * without need, but we won't read more data until later on
     * anyway. */
    if (ms->size - ms->woff == 0) {
      cp = (char *)ctxt->mcRealloc(ms->buf, ms->size * 2);
      if (cp == NULL) {
	/* No sense in continuing if we can't read(2) in a whole
	 * line. There are likely bigger problems on a system if its
	 * experiencing a problem like this. */
	MCM_ERR(MCM_ERR_MEM_REALLOC);
	return NULL;
      } else {
	/* Fix the other pointers */
	ms->buf = cp;
      }

      /* Note the change in buffer size */
      ms->size *= 2;
    }

    /* Check if we are at the end of a response (assume \r\n will
     * guarantee this.  This is safe for all types of commands except
     * data retrieval (handled in another function), and the stats
     * command.  I'm not too worried about the stats command as it
     * isn't likely to be used in production systems, or at least not
     * often.
     *
     * But it should be understood that this check may tell us we got
     * all the data when we actually have not in the event that we
     * read only part of a stats command response but just happen to
     * be at the end of a line. */
    if (memcmp(&ms->buf[ms->woff - MCM_CSTRLEN("\r\n")], "\r\n", MCM_CSTRLEN("\r\n")) == 0) {
      goto done_reading;
    }

    /* No acceptable end of command output was read, and there was no
     * error indicated.  Try to read more data we probably just didn't
     * get it all the first time through. */
    goto try_read;
  }

  done_reading:
  /* There should always be a newline in our response now.  Try to
   * find it so we can put a null in and return just the line to the
   * caller. */
  cp = memchr(&ms->buf[ms->startoff], (int)'\n', (size_t)ms->woff - ms->startoff);
  if (cp == NULL) {
    /* We suck.  Return null. */
    MCM_ERR_MSG(MCM_ERR_PROTO, "no \\n in server response");
    mcm_server_deactivate(ctxt, mc, ms);
    return NULL;
  }

#ifdef DEBUG_MC_PROTO_ASSERT
  /* At this point, we're guaranteed to have a complete line in the
   * buffer.  cp should be pointing to a newline character. */
  if (*cp != '\n') {
    MCM_ERR_MSG(MCM_ERR_PROTO, "expecting a \\n");
    mcm_server_deactivate(ctxt, mc, ms);
    return NULL;
  }

  /* Protocol check, make sure there's a carriage return before the
   * newline. */
  if (*(cp - 1) != '\r') {
    MCM_ERR_MSG(MCM_ERR_PROTO, "no \\r before the \\n");
    mcm_server_deactivate(ctxt, mc, ms);
    return NULL;
  }
#endif

  cur = &ms->buf[ms->startoff];
  ms->startoff += (size_t)(cp - &ms->buf[ms->startoff]) + 1; /* Advance to the start of the next line */

  /* Handy little debugging function: */
  /* MCM_ERR_MSG_LEN(MCM_ERR_LVL_INFO, cp, (size_t)(ms->start - cp - 2)); */
  return cur;
}


static void
mcm_retrieve_data(struct memcache_ctxt *ctxt, struct memcache_req *req,
		  struct memcache *mc, struct memcache_server *ms) {
  ssize_t rb;
  char *cp, *end;
  ssize_t len = 0;
  ssize_t bytes_read = 0;
  struct memcache_res *res = NULL;
  size_t bytes = 0;
  u_int16_t flags;
  int grow = 0;

  ms->soff = ms->woff = ms->startoff = 0;

  read_more:
  /* We just wrote something out and are switching to read(2)ing
   * a response, ensure we're using NBIO. */
  mcm_server_block(ctxt, ms, NBIO);
  mcm_server_readable(ctxt, ms, &ms->tv);
  rb = read(ms->fd, &ms->buf[ms->woff], ms->size - ms->woff);
  switch (rb) {
  case -1:
    /* We're in non-blocking mode, don't abort because of EAGAIN or
     * EINTR */
    switch (errno) {
    case EAGAIN:
    case EINTR:
      goto read_more;
    case ECONNRESET:
    case EINVAL:
      /* The server was restarted.  We shouldn't deactivate the
       * server, but we should try and reconnect to it and roll from
       * there. */
      MCM_ERR_MSG_LVL(MCM_ERR_SYS_READ, strerror(errno), MCM_ERR_LVL_INFO);

      mcm_server_disconnect(ctxt, ms);

      if (mcm_server_connect(ctxt, mc, ms) == -1) {
	mcm_server_deactivate(ctxt, mc, ms);
	ms = mcm_server_connect_next_avail(ctxt, mc, ms->_last_hash);

	if (ms == NULL)
	  return;
      } else {
	MCM_ERRX(MCM_ERR_MC_RECONN);
      }

      /* Need to send the last request again that way we have
       * something to read(2). */
      mcm_server_send_last_cmd(ctxt, mc, ms);

      /* Ok, now try the read(2) again.  In theory, we could loop
       * forever if the server keeps resetting the connection (ie,
       * firewall that doesn't drop packets but sends back an ICMP
       * port reset message).  *shrug* Fuck 'em/it. */
      goto read_more;
    case EBADF:
    case EFAULT:
      /* This shouldn't happen and if it does, we're pooched: better
       * dump. */
      MCM_ERR_MSG_LVL(MCM_ERR_SYS_READ, strerror(errno), MCM_ERR_LVL_FATAL);
      break;

    default:
      /* We shouldn't ever get here. */
      MCM_ERR_MSG(MCM_ERR_ASSERT, strerror(errno));
      return;
    }
  case 0:
    /* Seems like an error to me if the server closes its connection
     * here.  deactivate the server instead of just closing it. */
    mcm_server_deactivate(ctxt, mc, ms);
    MCM_ERR_MSG(MCM_ERR_SYS_READ, "server unexpectedly closed connection");
    return;
  default:
    ms->woff += rb;
  }

  if (ms->size - ms->woff == 0 || grow) {
    /* Store the difference between ms->woff and ms->buf so that
     * it can be restored after mcRealloc. */
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size * 2);
    if (cp == NULL) {
      /* No sense in continuing if we can't read(2) in a whole
       * line. There are likely bigger problems on a system if its
       * experiencing a problem like this. */
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return;
    } else {
      ms->buf = cp;
      ms->size *= 2;
      grow = 0;
    }
  }

next_value:
  /* We got some data, try to figure out what it is. */
  if (len == 0) {
    /* We haven't yet read a "header" line VALUE ... .. ..\r\n This
     * should be one. Try to match it. */
    end = memchr(&ms->buf[ms->startoff], (int)'\n', ms->woff - ms->startoff);
    if (end == NULL) {
      /* No \n yet... keep reading... */
      grow++;
      goto read_more;
    }

    if (memcmp(&ms->buf[ms->soff], "VALUE ", MCM_CSTRLEN("VALUE ")) == 0) {
      ms->soff += MCM_CSTRLEN("VALUE ");

      /* First assume the responses are in order as an optimization.
       * If this fails we will scan the whole list. */
      if (res != NULL && res->entries.tqe_next != NULL) {
	for (res = res->entries.tqe_next; res != NULL; res = res->entries.tqe_next) {
	  if (ms->woff - ms->soff >= res->len) {
	    if (memcmp(&ms->buf[ms->soff], res->key, res->len) == 0) {
	      break;
	    }
	  }
	}
      } else {
	for (res = req->query.tqh_first; res != NULL; res = res->entries.tqe_next) {
	  if (ms->woff - ms->soff >= res->len) {
	    if(memcmp(&ms->buf[ms->soff], res->key, res->len) == 0) {
	      break;
	    }
	  }
	}
      }

      if (res == NULL) {
	MCM_ERR_MSG(MCM_ERR_PROTO, "server sent data for key not in request");
	exit(1);
      }

      ms->soff += res->len;

      flags = (u_int16_t)strtol(&ms->buf[ms->soff], &end, 10);
      if (flags == 0 && ((errno == EINVAL && end == &ms->buf[ms->soff]) || errno == ERANGE)) {
	MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid flags");
	mcm_server_deactivate(ctxt, mc, ms);
	return;
      }
      res->flags = flags;
      ms->soff += end - &ms->buf[ms->soff];

      bytes = (size_t)strtol(&ms->buf[ms->soff], &end, 10);
      if (bytes == 0 && ((errno == EINVAL && end == &ms->buf[ms->soff]) || errno == ERANGE)) {
	MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid bytes");
	mcm_server_deactivate(ctxt, mc, ms);
	return;
      }
      res->bytes = bytes;
      ms->soff += end - &ms->buf[ms->soff];
      len = bytes;
      bytes_read = 0;

      if (ms->buf[ms->soff] == '\r' && ms->buf[ms->soff + 1] == '\n') {
	ms->soff += 2;
      } else {
	MCM_ERR(MCM_ERR_PROTO);
	mcm_server_deactivate(ctxt, mc, ms);
	return;
      }
    } else if (memcmp(&ms->buf[ms->soff], "END\r\n", MCM_CSTRLEN("END\r\n")) == 0) {
      return;
    } else {
      goto read_more;
    }
  }

  bytes_read = ms->woff - (ms->soff - ms->startoff);

  /* Check if we have read all the data the plus a \r\n (plus partial
   * next item or END) */
  if (bytes_read > len + 2) {
    res->_flags |= MCM_RES_FOUND;
    if (res->size == 0) {
      res->val = ctxt->mcMallocAtomic(res->bytes + 1);
      if (res->val == NULL) {
	MCM_ERR_MSG(MCM_ERR_ASSERT, "memory was not allocated for res->val");
	mcm_server_deactivate(ctxt, mc, ms);
	return;
      }
      ((char *)res->val)[res->bytes] = '\0';
      res->size = res->bytes;
    }

    if (res->size >= res->bytes) {
      /* We have room for the whole value.  Read it. */
      memcpy(res->val, &ms->buf[ms->soff], res->bytes);
      ms->soff += res->bytes;
    } else {
      /* Part of the value will be truncated. */
      memcpy(res->val, &ms->buf[ms->soff], res->size);
      ms->soff += res->size;
    }
    len = 0;
    bytes_read = 0;
    if (ms->buf[ms->soff] == '\r' && ms->buf[ms->soff + 1] =='\n') {
      ms->soff += 2;
    } else {
      MCM_ERR(MCM_ERR_PROTO);
      mcm_server_deactivate(ctxt, mc, ms);
      return;
    }

    if (ms->soff - ms->startoff >= MCM_CSTRLEN("END\r\n") && memcmp(&ms->buf[ms->soff], "END\r\n", MCM_CSTRLEN("END\r\n")) == 0) {
      /* Done reading data. */
      return;
    } else {
      goto next_value;
    }
  } else if (bytes_read <= len + 2) {
    goto read_more;
  }
}


#ifdef USE_CRC32_HASH
#include "crc32_table.h"
#endif /* USE_CRC32_HASH */


u_int32_t
mcm_hash_key(const struct memcache_ctxt *ctxt, const char *key, const size_t len) {
#ifdef USE_CRC32_HASH
  size_t i;
  u_int32_t crc;

  crc = ~0;

  for (i = 0; i < len; i++)
    crc = (crc >> 8) ^ crc32tab[(crc ^ (key[i])) & 0xff];

  return((~crc >> 16) & 0x7fff);
#else
# ifdef USE_PERL_HASH
  u_int32_t h, i;
  char *p;

  i = len;	/* Work back through the key length */
  p = key;	/* Character pointer */
  h = 0;	/* The hash value */

  while (i--) {
    h += *p++;
    h += (h << 10);
    h ^= (h >> 6);
  }
  h += (h << 3);
  h ^= (h >> 11);
  h += (h << 15);

  return h;
# else
#  ifdef USE_ELF_HASH
  u_int32_t g, h, i;
  char *p;

  i = len;	/* Work back through the key length */
  p = key;	/* Character pointer */
  h = 0;	/* The hash value */

  while (i--) {
    h = (h << 4) + *p++;
    if (g = h & 0xF0000000)
      h ^= g >> 24;
    h &= ~g;
  }

  return h;
#  else
#   error "Please choose USE_CRC32_HASH, USE_ELF_HASH, or USE_PERL_HASH as a hashing scheme when compiling memcache"
#  endif
# endif
#endif
}


u_int32_t
mcm_incr(struct memcache_ctxt *ctxt, struct memcache *mc,
	 const char *key, const size_t key_len, const u_int32_t val) {
  return mcm_atomic_cmd(ctxt, mc, str_incr_cmd, str_incr_len, key, key_len, val);
}


struct memcache *
mcm_new(struct memcache_ctxt *ctxt) {
  struct memcache *mc;

#ifdef USE_MC_IOVEC
  if (sizeof(struct iovec) != sizeof(struct iovec_old))
    MCM_ERR_MSG(MCM_ERR_ASSERT, "memcache(3)'s struct iovec is a different size than the OS's struct iovect.  Please recopmile without USE_MC_IOVEC.");
#endif

  mc = (struct memcache *)ctxt->mcMalloc(sizeof(struct memcache));
  if (mc != NULL) {
    bzero(mc, sizeof(struct memcache));

    TAILQ_INIT(&mc->server_list);

    /* Set any default values */
    mc->tv.tv_sec = 2;
    mc->tv.tv_usec = 600;
  }

  /* By default, ignore INFO and NOTICE level messages */
  ctxt->MCM_ERR_MASK = MCM_ERR_LVL_INFO | MCM_ERR_LVL_NOTICE;

  return mc;
}


#ifdef SEAN_HACKS
void
mcm_refresh(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_req *req) {
  mcm_fetch_cmd(ctxt, mc, req, str_refresh_cmd, str_refresh_len);
}
#endif


const u_int32_t
mcm_reldate(const struct memcache_ctxt *ctxt) {
  return MEMCACHE_RELDATE;
}


int
mcm_replace(struct memcache_ctxt *ctxt, struct memcache *mc,
	    const char *key, const size_t key_len,
	    const void *val, const size_t bytes,
	    const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(ctxt, mc, str_replace_cmd, str_replace_len, key, key_len, val, bytes, expire, flags);
}


struct memcache_res *
mcm_req_add(const struct memcache_ctxt *ctxt, struct memcache_req *req, const char *key, const size_t len) {
  struct memcache_res *res;
  res = mcm_res_new(ctxt);

  res->key = mcm_strdup(ctxt, key);
  res->_flags |= MCM_RES_NEED_FREE_KEY;
  res->len = len;

  TAILQ_INSERT_TAIL(&req->query, res, entries);
  req->num_keys++;

  return res;
}


struct memcache_res *
mcm_req_add_ref(const struct memcache_ctxt *ctxt, struct memcache_req *req, mc_const char *key, const size_t len) {
  struct memcache_res *res;
  res = mcm_res_new(ctxt);

  res->key = key;
  res->len = len;

  TAILQ_INSERT_TAIL(&req->query, res, entries);
  req->num_keys++;

  return res;
}


void
mcm_req_free(const struct memcache_ctxt *ctxt, struct memcache_req *req) {
  while (req->query.tqh_first != NULL)
    mcm_res_free(ctxt, req, req->query.tqh_first);

  while (req->cb.tqh_first != NULL)
    mcm_res_cb_free(req, req->cb.tqh_first);

  ctxt->mcFree(req);
}


struct memcache_req *
mcm_req_new(const struct memcache_ctxt *ctxt) {
  struct memcache_req *req;

  req = (struct memcache_req *)ctxt->mcMalloc(sizeof(struct memcache_req));
  if (req != NULL) {
    bzero(req, sizeof(struct memcache_req));

    TAILQ_INIT(&req->query);
    TAILQ_INIT(&req->cb);
  }

  return req;
}


int
mcm_res_attempted(const struct memcache_ctxt *ctxt,
		  const struct memcache_res *res) {
  return res->_flags & MCM_RES_ATTEMPTED ? 1 : 0;
}


int
mcm_res_found(const struct memcache_ctxt *ctxt,
	      const struct memcache_res *res) {
  return ((res->_flags & (MCM_RES_ATTEMPTED | MCM_RES_FOUND)) == (MCM_RES_ATTEMPTED | MCM_RES_FOUND) ? 1 : 0);
}


void
mcm_res_free(const struct memcache_ctxt *ctxt, struct memcache_req *req, struct memcache_res *res) {
  TAILQ_REMOVE(&req->query, res, entries);
  if ((res->_flags & MCM_RES_NEED_FREE_KEY) == MCM_RES_NEED_FREE_KEY)
    ctxt->mcFree((void *)res->key);

  if (((res->_flags & (MCM_RES_FREE_ON_DELETE | MCM_RES_NO_FREE_ON_DELETE)) ==
       (MCM_RES_FREE_ON_DELETE | MCM_RES_NO_FREE_ON_DELETE)) ||
      res->_flags & MCM_RES_FREE_ON_DELETE) {
    if (res->size > 0)
      ctxt->mcFree(res->val);
  }

  ctxt->mcFree(res);
}


void
mcm_res_free_on_delete(const struct memcache_ctxt *ctxt, struct memcache_res *res, const int fod) {
  if (fod) {
    res->_flags &= ~MCM_RES_NO_FREE_ON_DELETE;
    res->_flags |= MCM_RES_FREE_ON_DELETE;
  } else {
    res->_flags &= ~MCM_RES_FREE_ON_DELETE;
    res->_flags |= MCM_RES_NO_FREE_ON_DELETE;
  }
}


static struct memcache_res *
mcm_res_new(const struct memcache_ctxt *ctxt) {
  struct memcache_res *res;

  res = (struct memcache_res *)ctxt->mcMalloc(sizeof(struct memcache_res));
  if (res != NULL) {
    bzero(res, sizeof(struct memcache_res));

    /* Default values */
    res->_flags = MCM_RES_FREE_ON_DELETE | MCM_RES_NO_FREE_ON_DELETE; /* unset */
  }

  return res;
}


static void
mcm_res_cb_free(struct memcache_req *req, struct memcache_res_cb *cb) {
  mcFreeFunc freeFunc;

  if (cb == NULL || cb->ctxt == NULL)
    return;

  TAILQ_REMOVE(&req->cb, cb, entries);
  freeFunc = cb->ctxt->mcFree;
  (freeFunc)(cb);
}


static struct memcache_res_cb *
mcm_res_cb_new(const struct memcache_ctxt *ctxt) {
  struct memcache_res_cb *cb;

  cb = (struct memcache_res_cb *)ctxt->mcMalloc(sizeof(struct memcache_res_cb));
  if (cb != NULL) {
    bzero(cb, sizeof(struct memcache_res_cb));
  }

  return cb;
}


int
mcm_res_register_fetch_cb(struct memcache_ctxt *ctxt, struct memcache_req *req,
			  struct memcache_res *res, mcResCallback callback, void *misc) {
  struct memcache_res_cb *cb;

  if (callback == NULL || req == NULL || res == NULL || ctxt == NULL)
    return -1;

  cb = mcm_res_cb_new(ctxt);
  if (cb == NULL)
    return -2;

  cb->ctxt = ctxt;
  cb->req = req;
  cb->cb = callback;
  cb->res = res;
  cb->misc = misc;

  TAILQ_INSERT_TAIL(&req->cb, cb, entries);

  return 0;
}


static void
mcm_reset_buf(struct memcache_server *ms) {
  /* Reset our offsets */
  ms->soff = ms->startoff = 0;
  ms->woff = -1;
}


/* SHIT!  This is a fucked function.  It needs to re-inject the server
 * into the server list.  The dealings with down servers and
 * reactivaging them in server lists needs to be thought out quite a
 * bit more.  XXX */
int
mcm_server_activate(const struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  switch (ms->active) {
  case 'd':
    ms->active = 'u';
    mc->live_servers[mc->num_live_servers] = ms;
    mc->live_servers++;

    return 0;
  case 'n':
    MCM_ERR_MSG(MCM_ERR_ASSERT, "unable to activate a server that does not exist");
    return -1;
  case 't':
    MCM_ERR_MSG_LVL(MCM_ERR_ASSERT, "unable to activate a server that hasn't been added to the server list", MCM_ERR_LVL_INFO);
    return -2;
  case 'u':
    MCM_ERR_MSG_LVL(MCM_ERR_ASSERT, "unable to activate a server that is active", MCM_ERR_LVL_INFO);
    return -3;
  default:
    MCM_ERR(MCM_ERR_ASSERT);
  }

  /* MCM_ERR(MCM_ERR_ASSERT) should catch this case */
  return 0;
}


int
mcm_server_activate_all(const struct memcache_ctxt *ctxt, struct memcache *mc) {
  struct memcache_server *ms;

  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next) {
    if (ms->active == 'd')
      mcm_server_activate(ctxt, mc, ms);
  }

  return 0;
}


int
mcm_server_add(struct memcache_ctxt *ctxt, struct memcache *mc, const char *hostname, const char *port) {
  return mcm_server_add2(ctxt, mc, hostname, strlen(hostname), port, strlen(port));
}


int
mcm_server_add2(struct memcache_ctxt *ctxt, struct memcache *mc, const char *hostname,
		const size_t hostname_len, const char *port, const size_t port_len) {
  struct memcache_server *ms;

  ms = mcm_server_new(ctxt);
  if (ms == NULL)
    return -1;

  if (hostname == NULL || hostname_len == 0) {
    ms->hostname = mcm_strdup(ctxt, "localhost");
  } else {
    ms->hostname = ctxt->mcMallocAtomic(hostname_len + 1);
    memcpy(ms->hostname, hostname, hostname_len);
    ms->hostname[hostname_len] = '\0';
  }

  if (ms->hostname == NULL) {
    mcm_server_free(ctxt, ms);
    return -2;
  }


  if (port == NULL || port_len == 0) {
    ms->port = mcm_strdup(ctxt, "11211");
  } else {
    ms->port = ctxt->mcMallocAtomic(port_len + 1);
    memcpy(ms->port, port, port_len);
    ms->port[port_len] = '\0';
  }

  if (ms->port == NULL) {
    mcm_server_free(ctxt, ms);
    return -3;
  }

  return mcm_server_add3(ctxt, mc, ms);
}


int
mcm_server_add3(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  int ret;

  ret = mcm_server_resolve(ctxt, ms);
  if (ret != 0) {
    MCM_ERR_MSG(MCM_ERR_NET_HOST, gai_strerror(ret));
    mcm_server_free(ctxt, ms);
    return -4;
  }

  /* Defaults from mc */
  if (ms->tv.tv_sec == 0 && ms->tv.tv_usec == 0) {
    ms->tv.tv_sec = mc->tv.tv_sec;
    ms->tv.tv_usec = mc->tv.tv_usec;
  }

  TAILQ_INSERT_TAIL(&mc->server_list, ms, entries);

  /* Add ms to the array of servers to try */
  if (mc->live_servers == NULL) {
    mc->num_live_servers = 1;
    mc->live_servers = (struct memcache_server**)ctxt->mcMalloc(sizeof(struct memcache_server*) * mc->num_live_servers);
    mc->live_servers[0] = ms;
  } else {
    /* Reallocate mc->live_servers to fit the number of struct
     * memcache_servers entries. */
    mc->num_live_servers++;
    mc->live_servers = (struct memcache_server**)ctxt->mcRealloc(mc->live_servers, sizeof(struct memcache_server*) * mc->num_live_servers);
    if (mc->live_servers == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      mcm_server_free(ctxt, ms);
      return -5;
    }

    /* Add the new server to the end of the list */
    mc->live_servers[mc->num_live_servers - 1] = ms;
  }

  return 0;
}


int
mcm_server_add4(struct memcache_ctxt *ctxt, struct memcache *mc, mc_const char *hostport) {
  return mcm_server_add5(ctxt, mc, hostport, strlen(hostport));
}


int
mcm_server_add5(struct memcache_ctxt *ctxt, struct memcache *mc,
		mc_const char *hostport, const size_t hostlen) {
  struct memcache_server *ms;
  char *cp;

  ms = mcm_server_new(ctxt);
  if (ms == NULL)
    return -1;

  /* Tease out the hostname and portname from a string that we expect
   * to look like "host:port". */
  if (hostport == NULL || hostlen == 0) {
    ms->hostname = mcm_strdup(ctxt, "localhost");
    if (ms->hostname == NULL) {
      mcm_server_free(ctxt, ms);
      return -2;
    }

    ms->port = mcm_strdup(ctxt, "11211");
    if (ms->port == NULL) {
      mcm_server_free(ctxt, ms);
      return -3;
    }
  } else {
    cp = mcm_strnchr(ctxt, hostport, ':', hostlen);
    if (*cp == '\0') {
      ms->hostname = mcm_strndup(ctxt, hostport, hostlen);
      if (ms->hostname == NULL) {
	mcm_server_free(ctxt, ms);
	return -2;
      }

      ms->port = mcm_strdup(ctxt, "11211");
      if (ms->port == NULL) {
	mcm_server_free(ctxt, ms);
	return -3;
      }
    } else {
      ms->hostname = mcm_strndup(ctxt, hostport, (size_t)(cp - hostport));
      if (ms->hostname == NULL) {
	mcm_server_free(ctxt, ms);
	return -2;
      }

      /* advance past the ':' and copy whatever is left as the port */
      cp++;
      ms->port = mcm_strndup(ctxt, cp, hostlen - (size_t)(cp - hostport));
      if (ms->port == NULL) {
	mcm_server_free(ctxt, ms);
	return -3;
      }
    }
  }

  return mcm_server_add3(ctxt, mc, ms);
}


static void
mcm_server_block(const struct memcache_ctxt *ctxt, struct memcache_server *ms, const int use_nbio) {
  if (ms->flags == -1) {
    ms->flags = fcntl(ms->fd, F_GETFL, 0);
    if (ms->flags == -1) {
      MCM_ERR_MSG(MCM_ERR_SYS_FCNTL, "fcntl(F_GETFL)");
      return;
    }
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      tml = asprintf(&tm, "ms(0x%p)'s fcntl(F_GETFL)'s fd flags: 0x%x", (void*)ms, ms->flags);
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif
  }

#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "ms(0x%p)->flags == 0x%x, ms->fd == %d, use_nbio == %d", (void*)ms, ms->flags, ms->fd, use_nbio);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif
  if (use_nbio != BLOCK) {
    /* Switch to non-blocking io */
    if ((ms->flags & O_NONBLOCK) != O_NONBLOCK) {
#ifdef DEBUG_MC_PROTO
      MCM_WARNX_MSG(MCM_ERR_TRACE, "Added O_NONBLOCK to flags");
#endif
      ms->flags |= O_NONBLOCK;
      goto set_flags;
#ifdef DEBUG_MC_PROTO
    } else {
      MCM_WARNX_MSG(MCM_ERR_TRACE, "Flags already have O_NONBLOCK bit");
#endif
    }
  } else {
    if ((ms->flags & O_NONBLOCK) == O_NONBLOCK) {
#ifdef DEBUG_MC_PROTO
      MCM_WARNX_MSG(MCM_ERR_TRACE, "Cleared O_NONBLOCK from flags");
#endif
      ms->flags &= ~O_NONBLOCK;
      goto set_flags;
#ifdef DEBUG_MC_PROTO
    } else {
      MCM_WARNX_MSG(MCM_ERR_TRACE, "Flags already missing O_NONBLOCK bit");
#endif
    }
  }
  return;

  set_flags:
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      tml = asprintf(&tm, "ms(0x%p)'s fcntl(F_SETFL)'s new fd flags: 0x%x", (void*)ms, ms->flags);
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif
  if (fcntl(ms->fd, F_SETFL, ms->flags) < 0) {
    MCM_ERR_MSG(MCM_ERR_SYS_FCNTL, "fcntl(F_SETFL)");
    return;
  }
#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "confirmed ms(0x%p)'s fcntl(F_GETFL)'s fd flags: 0x%x:0x%x", (void*)ms, fcntl(ms->fd, F_GETFL, 0), ms->flags);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
    }
  } while (0);
#endif
}


static int
mcm_server_connect(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  struct addrinfo *res;
  int i;
#ifdef TCP_NODELAY
  int val;
#endif
#ifdef HAVE_SELECT
  int ret;
#endif

  if (ms->fd != -1)
    return ms->fd;

#ifdef DEBUG_MC_PROTO
  MCM_WARNX_MSG(MCM_ERR_TRACE, "Begin connect(2)");
#endif

  if (ms->active == 'd' || ms->active == 'n')
    return -1;

  if (ms->hostinfo == NULL || ms->hostinfo->ai_addrlen == 0) {
    i = mcm_server_resolve(ctxt, ms);
    if (i != 0) {
      MCM_ERR_MSG(MCM_ERR_NET_HOST, gai_strerror(i));
      ms->active = 'n';
      return -1;
    }
  }

  for (i = 0, res = ms->hostinfo; res != NULL; res = res->ai_next, i++) {
    ms->fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (ms->fd < 0) {
#ifdef AF_INET6
      if (errno == EPROTONOSUPPORT && res->ai_family == AF_INET6)
	continue;
#endif
      MCM_ERR(MCM_ERR_SYS_SOCKET);
      continue;
    }

    /* Set the file descriptor to be non-blocking. */
    mcm_server_block(ctxt, ms, NBIO);

#ifdef TCP_NODELAY
    val = 1;
    if (setsockopt(ms->fd, IPPROTO_TCP, TCP_NODELAY, &val, (socklen_t)sizeof(val)) != 0) {
      MCM_WARN_MSG(MCM_ERR_SYS_SETSOCKOPT, "setsockopt(TCP_NODELAY) failed");
    }
#endif

    /* If any of the setsockopt(2) calls fail, close the socket, set
     * the file descriptor to -1, and continue trying to connect to
     * the rest of the servers that match this hostname.  More than
     * likely there is only one IP per host name, but, in the event
     * there isn't, continue to the next entry. */
    if (setsockopt(ms->fd, SOL_SOCKET, SO_SNDTIMEO, &ms->tv, (socklen_t)sizeof(struct timeval)) != 0) {
      MCM_ERR_MSG(MCM_ERR_SYS_SETSOCKOPT, "setsockopt(SO_SNDTIMEO) failed");
      mcm_server_disconnect(ctxt, ms);
      continue;
    }

    if (setsockopt(ms->fd, SOL_SOCKET, SO_RCVTIMEO, &ms->tv, (socklen_t)sizeof(struct timeval)) != 0) {
      MCM_ERR_MSG(MCM_ERR_SYS_SETSOCKOPT, "setsockopt(SO_RCVTIMEO) failed");
      mcm_server_disconnect(ctxt, ms);
      continue;
    }

  get_conn_status:
    ret = connect(ms->fd, res->ai_addr, (socklen_t)res->ai_addrlen);
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      char ch[256], cs[256];

      tml = getnameinfo(res->ai_addr, res->ai_addr->sa_len, ch, sizeof(ch), cs, sizeof(cs), NI_NUMERICHOST | NI_NUMERICSERV);
      if (tml != 0) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, "Unable to get address");
	return -1;
      }

      tml = asprintf(&tm, "connect(2) return status for %s:%s: %d", ch, cs, ret);
      if (tml > 0 && tm != NULL) {
	MCM_WARN_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif
    if (ret == 0) {
      return ms->fd;
    } else {
      /* Leave error handling in the event I figure out why connect(2)
       * and select(2) weren't returning when polling for a writable
       * file descriptor. */
      switch (errno) {
      case EISCONN:
	return ms->fd;
      case EINPROGRESS:
	ret = mcm_server_writable(ctxt, ms, &ms->tv);
	if (ret == -1) {
	  MCM_ERR_MSG_LVL(MCM_ERR_SYS_SELECT, "select(2) failed for writable status", MCM_ERR_LVL_WARN);
	} else if (ret == 0) {
	  MCM_ERR_MSG_LVL(MCM_ERR_SYS_SELECT, "select(2) timed out on establishing connection", MCM_ERR_LVL_WARN);
	  printf("connect(): %d\n", connect(ms->fd, res->ai_addr, (socklen_t)res->ai_addrlen));
	} else {
	  goto get_conn_status;
	}
	/* Fall through */
      default:
	MCM_ERR(MCM_ERR_SYS_CONNECT);
	mcm_server_disconnect(ctxt, ms);
	continue;
      }
    }
  }

  /* If none of the IP addresses for this hostname work, remove the
   * server from the live_server list (we assume they're live by
   * default) and return -1. */
  mcm_server_deactivate(ctxt, mc, ms);
  return -1;
}


static struct memcache_server *
mcm_server_connect_next_avail(struct memcache_ctxt *ctxt, struct memcache *mc, const u_int32_t hash) {
  struct memcache_server *ms, *nms;

  if (mc->num_live_servers == 0) {
    MCM_ERRX(MCM_ERR_MC_SERV_LIST);
    return NULL;
  }

  ms = mcm_server_find(ctxt, mc, hash);
  if (ms == NULL) {
    MCM_ERRX(MCM_ERR_MC_VALID_SERVER);
    return NULL;
  }

  while (mcm_server_connect(ctxt, mc, ms) == -1) {
    MCM_ERR(MCM_ERR_NET_CONNECT);
    mcm_server_deactivate(ctxt, mc, ms);

    nms = mcm_server_find(ctxt, mc, ms->_last_hash);
    if (nms == NULL) {
      MCM_ERRX(MCM_ERR_MC_SERV_LIST);
      return NULL;
    }

    nms->_last_cmd = ms->_last_cmd;
    nms->_last_cmd_size = ms->_last_cmd_size;
    ms = nms;
  }

  /* If there was a present left behind by the last memcache_server,
   * assume ownership of the command. */
  if (ctxt->_last_cmd != NULL) {
    ms->_last_cmd = ctxt->_last_cmd;
    ms->_last_cmd_size = ctxt->_last_cmd_size;
    ms->_last_hash = ctxt->_last_hash;
    ctxt->_last_cmd = NULL;
    ctxt->_last_cmd_size = 0;
    ctxt->_last_hash = 0;
  }

  return ms;
}


void
mcm_server_deactivate(struct memcache_ctxt *ctxt, struct memcache *mc,
		      struct memcache_server *ms) {
  u_int32_t i, found;

  /* Since adding servers is so rare, and servers do come back, don't
   * bother mcRealloc(3)'ing mc->live_servers.  Instead, just find the
   * dead server in the array, remove it, shift the remaining servers
   * down a place in the array, and decrement number of servers in the
   * array.  This should making adding servers back to the list more
   * efficient. */
  for (i = 0, found = 0; i < mc->num_live_servers; i++) {
    if (found == 0) {
      if (mc->live_servers[i] == ms)
	found = 1;
      else
	continue;
    }

    mc->live_servers[i] = mc->live_servers[i + 1];
  }

  if (found != 0) {
    mc->num_live_servers--;
    /* If we've still got the server marked up, then set it to down.
     * If we've already marked it as down, keep the original reason
     * since it's going to be more verbose than saying 'down'.  A
     * 'down' server can be resurrected, where as a server that's
     * marked 'no host' won't ever be resurrected. */
    if (ms->active == 'u')
      ms->active = 'd';

    mcm_server_disconnect(ctxt, ms);
  }

  /* Stash this server's command in the context */
  ctxt->_last_cmd = ms->_last_cmd;
  ctxt->_last_cmd_size = ms->_last_cmd_size;
  ctxt->_last_hash = ms->_last_hash;
}


void
mcm_server_disconnect(const struct memcache_ctxt *ctxt, struct memcache_server *ms) {
#ifdef DEBUG_MC_PROTO
    MCM_WARNX_MSG(MCM_ERR_TRACE, "Attempting to disconnect");
#endif
  if (ms->fd != -1) {
#ifdef DEBUG_MC_PROTO
    MCM_WARNX_MSG(MCM_ERR_TRACE, "Closing file descriptor");
#endif
    if (close(ms->fd) != 0)
      MCM_ERR(MCM_ERR_SYS_CLOSE);
    mcm_server_init(ctxt, ms);
  }
}


void
mcm_server_disconnect_all(const struct memcache_ctxt *ctxt, const struct memcache *mc) {
  struct memcache_server *ms;

  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next)
    mcm_server_disconnect(ctxt, ms);
}


struct memcache_server *
mcm_server_find(const struct memcache_ctxt *ctxt, struct memcache *mc, const u_int32_t hash) {
  struct memcache_server *ms;

  if (mc->num_live_servers < 1)
    return NULL;

  /* Grab the correct server from the list. */
  ms = mc->live_servers[hash % mc->num_live_servers];

  /* Store the last hash value used to find this server.  In the event
   * that this server dies, we use this value to automatically fall
   * back to the next server. */
  ms->_last_hash = hash;

  return ms;
}


void
mcm_server_free(const struct memcache_ctxt *ctxt, struct memcache_server *ms) {
  if (ms == NULL)
    return;

  if (ms->hostinfo != NULL)
    freeaddrinfo(ms->hostinfo);

  if (ms->hostname != NULL)
    ctxt->mcFree(ms->hostname);

  if (ms->port != NULL)
    ctxt->mcFree(ms->port);

  if (ms->buf != NULL)
    ctxt->mcFree(ms->buf);

  mcm_server_disconnect(ctxt, ms);

  ctxt->mcFree(ms);
}


static void
mcm_server_init(const struct memcache_ctxt *ctxt, struct memcache_server *ms) {
  ms->active = 't';
  ms->fd = -1;
  ms->flags = -1;
  if (ms->size == 0)
    ms->size = MAX(INIT_GET_BUF_SIZE, 1);
  ms->startoff = ms->soff = 0;
  ms->woff = -1;
}


struct memcache_server *
mcm_server_new(const struct memcache_ctxt *ctxt) {
  struct memcache_server *ms;

  ms = (struct memcache_server *)ctxt->mcMalloc(sizeof(struct memcache_server));
  if (ms != NULL) {
    bzero(ms, sizeof(struct memcache_server));

    /* Set default values */
    mcm_server_init(ctxt, ms);

    ms->buf = (char *)ctxt->mcMallocAtomic(ms->size);
    if (ms->buf == NULL) {
      ctxt->mcFree(ms);
      return NULL;
    }
  }

  return ms;
}


static int
mcm_server_readable(struct memcache_ctxt *ctxt, struct memcache_server *ms, struct timeval *tv) {
#ifdef HAVE_SELECT
  int ret;

  FD_ZERO(&ms->fds);
  FD_SET(ms->fd, &ms->fds);

#ifdef DEBUG_MC_PROTO_ASSERT
  if (FD_ISSET(ms->fd, &ms->fds) == 0) {
    MCM_ERRX(MCM_ERR_ASSERT);
    return -1;
  }
#endif

#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "Begin select(2)'ing to see if fd %d is read(2)able.  Timeout %lu.%lu", ms->fd, (u_int64_t)tv->tv_sec, (u_int64_t)tv->tv_usec);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif

  /* Before we read(2) anything, check to make sure there is data
   * available to be read(2).  No sense in wastefully calling
   * read(2) constantly in a loop. */
  ret = select(ms->fd + 1, &ms->fds, NULL, NULL, tv);
#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "fd's ready to be read(2): %d/%d", ret, FD_SETSIZE);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif
  switch (ret) {
  case -1:
    switch (errno) {
    case EINTR:
      return 0;
    default:
      MCM_ERR(MCM_ERR_SYS_SELECT);
      break;
    }
  case 0:
    MCM_ERR_MSG(MCM_ERR_TIMEOUT, "select(2) call timed out for read(2)able fds");
    break;
  }

#ifdef DEBUG_MC_PROTO
  MCM_WARNX_MSG(MCM_ERR_TRACE, "End select(2)'ing to see if fd is read(2)able");
#endif

  return ret;
#else
  return 1;
#endif
}


static int
mcm_server_resolve(struct memcache_ctxt *ctxt, struct memcache_server *ms) {
  struct addrinfo hints, *res;
  int ret;

#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "Resolving the host %s:%s", ms->hostname, ms->port);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif

  /* Resolve the hostname ahead of time */
  bzero(&hints, sizeof(struct addrinfo));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_protocol = IPPROTO_TCP;
  ret = getaddrinfo(ms->hostname, ms->port, &hints, &ms->hostinfo);
  if (ret != 0) {
#ifdef DEBUG_MC_PROTO
    do {
      char *tm;
      size_t tml;
      tml = asprintf(&tm, "getaddrinfo(): %s", gai_strerror(ret));
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif

    return ret;
  }

  for (res = ms->hostinfo; res != NULL; res = res->ai_next) {
    ms->num_addrs++;
#ifdef DEBUG_MC_PROTO
    do {
      char *tm, ch[256], cs[256];
      size_t tml;
      tml = getnameinfo(res->ai_addr, res->ai_addr->sa_len, ch, sizeof(ch), cs, sizeof(cs), NI_NUMERICHOST | NI_NUMERICSERV);
      if (tml != 0) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, "Unable to get address");
	return tml;
      }

      tml = asprintf(&tm, "Resolved host to \"%s:%s\"", ch, cs);
      if (tml > 0 && tm != NULL) {
	MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
	free(tm);
      }
    } while (0);
#endif

  }

  return 0;
}


static ssize_t
mcm_server_send_cmd(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms,
		    struct iovec *cmd, size_t cmd_len) {
  ssize_t ret;

  /* Record the last command written */
  ms->_last_cmd = cmd;
  ms->_last_cmd_size = cmd_len;

#ifdef DEBUG_MC_PROTO
  MCM_WARNX_MSG(MCM_ERR_TRACE, "Sending the following data to the server:");
  writev(STDERR_FILENO, (iovec_cast)cmd, cmd_len);
#endif

  write_again:
  mcm_server_block(ctxt, ms, BLOCK);
  ret = writev(ms->fd, (iovec_cast)cmd, cmd_len);
#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "%d = writev(2), errno = %d", (int)ret, errno);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif
  if (ret < 0) {
    switch (errno) {
    case EAGAIN:
    case EINTR:
      /* This shouldn't happen since we're in non-blocking IO, but,
       * since I'm headed down the road of using non-blocking IO to
       * blast out requests to multiple servers at the same time, add
       * a goto. */
      goto write_again;
    case EBADF:
    case EDESTADDRREQ:
      /* Need to reconnect */
      MCM_ERR_MSG_LVL(MCM_ERR_MC_SEND_CMD, strerror(errno), MCM_ERR_LVL_INFO);
      mcm_server_disconnect(ctxt, ms);

      ms = mcm_server_connect_next_avail(ctxt, mc, ms->_last_hash);
      goto write_again;
      break;
    case ENOBUFS:
      /* Hrm... we can try again, but if an OS has run out of buffers,
       * then they're going to be hurting for reasons beyond the
       * control of memcache(3).  Still, to play nice, return 0. */
      return 0;
    case EDQUOT:
    case EFAULT:
    case EFBIG:
    case EINVAL:
    case EIO:
    case ENOSPC:
    case EPIPE:
    default:
      MCM_ERR_MSG_LVL(MCM_ERR_MC_SEND_CMD, strerror(errno), MCM_ERR_LVL_FATAL);
      mcm_server_deactivate(ctxt, mc, ms);
      /* If we're here, the game's up and we can't continue. */
      return 0;
    }
  } else {
    return ret;
  }
}


inline static ssize_t
mcm_server_send_last_cmd(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  return mcm_server_send_cmd(ctxt, mc, ms, ms->_last_cmd, ms->_last_cmd_size);
}


struct memcache_server_stats *
mcm_server_stats(struct memcache_ctxt *ctxt, struct memcache *mc, struct memcache_server *ms) {
  struct memcache_server_stats *s;
  char *cp, *cur;
  struct iovec sv[1];

  if (mcm_server_connect(ctxt, mc, ms) == -1)
    return NULL;

  sv[0].iov_base = "stats\r\n";
  sv[0].iov_len = MCM_CSTRLEN("stats\r\n");
  if (mcm_server_send_cmd(ctxt, mc, ms, sv, 1) < 0)
    return NULL;

  s = mcm_server_stats_new(ctxt);
  if (s == NULL)
    return NULL;

  /* Reset our buffer */
  mcm_reset_buf(ms);

  for(;;) {
    cur = mcm_get_line(ctxt, mc, ms);

    if (cur != NULL && memcmp(cur, "STAT ", MCM_CSTRLEN("STAT ")) == 0) {
      cur = &cur[MCM_CSTRLEN("STAT ")];

      /* Time to loop through the potential stats keys.  Joy.  This is
       * going to complete in O(1 + 2 + 3 ... N) operations (currently


       * 190).  Ugh.  Don't know of a better way to handle this
       * without a hash.  Besides, this is just stats. */
      if (memcmp(cur, "pid ", MCM_CSTRLEN("pid ")) == 0) {
	cur = &cur[MCM_CSTRLEN("pid ")];
	s->pid = (pid_t)strtol(cur, &cp, 10);
	if (s->pid == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid pid");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "uptime ", MCM_CSTRLEN("uptime ")) == 0) {
	cur = &cur[MCM_CSTRLEN("uptime ")];
	s->uptime = (time_t)strtol(cur, &cp, 10);
	if (s->uptime == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid uptime");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "time ", MCM_CSTRLEN("time ")) == 0) {
	cur = &cur[MCM_CSTRLEN("time ")];
	s->time = (time_t)strtol(cur, &cp, 10);
	if (s->time == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid time");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "version ", MCM_CSTRLEN("version ")) == 0) {
	cur = &cur[MCM_CSTRLEN("version ")];
	for (cp = cur; !isspace(*cp); cp++);
	s->version = (char *)ctxt->mcMallocAtomic((size_t)(cp - cur + 1));
	if (s->version == NULL) {
	  MCM_ERR(MCM_ERR_MEM_MALLOC);
	} else {
	  memcpy(s->version, cur, (size_t)(cp - cur));
	  s->version[(size_t)(cp - cur)] = '\0';
	}
      } else if (memcmp(cur, "rusage_user ", MCM_CSTRLEN("rusage_user ")) == 0) {
	cur = &cur[MCM_CSTRLEN("rusage_user ")];
	s->rusage_user.tv_sec = (int32_t)strtol(cur, &cp, 10);
	if (s->rusage_user.tv_sec == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid rusage_user seconds");
	  return NULL;
	} else {
	  cur = cp;
#ifdef DEBUG_MC_PROTO_ASSERT
	  if (!(*cur == '.' || *cur == ':'))
	    MCM_WARNX_MSG(MCM_ERR_PROTO, "invalid separator");
	  else {
#endif
	    cur += 1; /* advance past colon */
	    s->rusage_user.tv_usec = (int32_t)strtol(cur, &cp, 10);
	    if (s->rusage_user.tv_usec == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	      MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid rusage_user microseconds");
	      return NULL;
	    } else {
	      cur = cp;
	    }
#ifdef DEBUG_MC_PROTO_ASSERT
	  }
#endif
	}

      } else if (memcmp(cur, "rusage_system ", MCM_CSTRLEN("rusage_system ")) == 0) {
	cur = &cur[MCM_CSTRLEN("rusage_system ")];
	s->rusage_system.tv_sec = (int32_t)strtol(cur, &cp, 10);
	if (s->rusage_system.tv_sec == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid rusage_system seconds");
	  return NULL;
	} else {
	  cur = cp;
#ifdef DEBUG_MC_PROTO_ASSERT
	  if (!(*cur == '.' || *cur == ':')) {
	    MCM_ERR_MSG(MCM_ERR_PROTO, "invalid separator");
	    return NULL;
	  } else {
#endif
	    cur += 1; /* advance past colon */
	    s->rusage_system.tv_usec = (int32_t)strtol(cur, &cp, 10);
	    if (s->rusage_system.tv_usec == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	      MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid rusage_system microseconds");
	      return NULL;
	    } else {
	      cur = cp;
	    }
#ifdef DEBUG_MC_PROTO_ASSERT
	  }
#endif
	}
      } else if (memcmp(cur, "curr_items ", MCM_CSTRLEN("curr_items ")) == 0) {
	cur = &cur[MCM_CSTRLEN("curr_items ")];
	s->curr_items = (u_int32_t)strtol(cur, &cp, 10);
	if (s->curr_items == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid curr_items");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "total_items ", MCM_CSTRLEN("total_items ")) == 0) {
	cur = &cur[MCM_CSTRLEN("total_items ")];
	s->total_items = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->total_items == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid total_items");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "bytes ", MCM_CSTRLEN("bytes ")) == 0) {
	cur = &cur[MCM_CSTRLEN("bytes")];
	s->bytes = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid bytes");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "curr_connections ", MCM_CSTRLEN("curr_connections ")) == 0) {
	cur = &cur[MCM_CSTRLEN("curr_connections ")];
	s->curr_connections = (u_int32_t)strtol(cur, &cp, 10);
	if (s->curr_connections == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid curr_connections");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "total_connections ", MCM_CSTRLEN("total_connections ")) == 0) {
	cur = &cur[MCM_CSTRLEN("total_connections ")];
	s->total_connections = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->total_connections == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid total_connections");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "connection_structures ", MCM_CSTRLEN("connection_structures ")) == 0) {
	cur = &cur[MCM_CSTRLEN("connection_structures ")];
	s->connection_structures = (u_int32_t)strtol(cur, &cp, 10);
	if (s->connection_structures == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid connection_structures");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "cmd_get ", MCM_CSTRLEN("cmd_get ")) == 0) {
	cur = &cur[MCM_CSTRLEN("cmd_get ")];
	s->cmd_get = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->cmd_get == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid cmd_get");
	  return NULL;
	} else {
	  cur = cp;
	}
#ifdef SEAN_HACKS
      } else if (memcmp(cur, "cmd_refresh ", MCM_CSTRLEN("cmd_refresh ")) == 0) {
	cur = &cur[MCM_CSTRLEN("cmd_refresh ")];
	s->cmd_refresh = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->cmd_refresh == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid cmd_refresh");
	  return NULL;
	} else {
	  cur = cp;
	}
#endif
      } else if (memcmp(cur, "cmd_set ", MCM_CSTRLEN("cmd_set ")) == 0) {
	cur = &cur[MCM_CSTRLEN("cmd_set ")];
	s->cmd_set = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->cmd_set == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid cmd_set");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "get_hits ", MCM_CSTRLEN("get_hits ")) == 0) {
	cur = &cur[MCM_CSTRLEN("get_hits ")];
	s->get_hits = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->get_hits == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid get_hits");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "get_misses ", MCM_CSTRLEN("get_misses ")) == 0) {
	cur = &cur[MCM_CSTRLEN("get_misses ")];
	s->get_misses = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->get_misses == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid get_misses");
	  return NULL;
	} else {
	  cur = cp;
	}
#ifdef SEAN_HACKS
      } else if (memcmp(cur, "refresh_hits ", MCM_CSTRLEN("refresh_hits ")) == 0) {
	cur = &cur[MCM_CSTRLEN("refresh_hits ")];
	s->refresh_hits = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->refresh_hits == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid refresh_hits");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "refresh_misses ", MCM_CSTRLEN("refresh_misses ")) == 0) {
	cur = &cur[MCM_CSTRLEN("refresh_misses ")];
	s->refresh_misses = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->refresh_misses == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid refresh_misses");
	  return NULL;
	} else {
	  cur = cp;
	}
#endif
      } else if (memcmp(cur, "bytes_read ", MCM_CSTRLEN("bytes_read ")) == 0) {
	cur = &cur[MCM_CSTRLEN("bytes_read ")];
	s->bytes_read = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes_read == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid bytes_read");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "bytes_written ", MCM_CSTRLEN("bytes_written ")) == 0) {
	cur = &cur[MCM_CSTRLEN("bytes_written ")];
	s->bytes_written = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->bytes_written == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid bytes_written");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else if (memcmp(cur, "limit_maxbytes ", MCM_CSTRLEN("limit_maxbytes ")) == 0) {
	cur = &cur[MCM_CSTRLEN("limit_maxbytes ")];
	s->limit_maxbytes = (u_int64_t)strtoll(cur, &cp, 10);
	if (s->limit_maxbytes == 0 && ((errno == EINVAL && cp == cur) || errno == ERANGE)) {
	  MCM_ERR_MSG(MCM_ERR_LIB_STRTOL, "invalid limit_maxbytes");
	  return NULL;
	} else {
	  cur = cp;
	}
      } else {
	for (cp = cur; !isspace(*cp); cp++);
	MCM_WARNX_MSGLEN(MCM_ERR_UNKNOWN_STAT, cur, (int)(cp - cur));
	return NULL;
      }

      /* Now that we've sucked in our stats value, set our cursor to
       * the end of the value. */
      cp = memchr(cur, (int)'\r', ms->woff - (cur - ms->buf));
      if (cp != NULL && cp[1] != '\n') {
	MCM_ERR_MSG(MCM_ERR_PROTO, "anticipated end of stats value: not at end of stats value");
	mcm_server_stats_free(ctxt, s);
	mcm_server_deactivate(ctxt, mc, ms);
	return NULL;
      }
    } else if (cur != NULL && memcmp(cur, "END", MCM_CSTRLEN("END")) == 0) {
      /* We're done reading in stats. */
      break;
    } else {
      MCM_ERR_MSGX(MCM_ERR_PROTO, "unable to handle response");
      return NULL;
    }
  }

  return s;
}


void
mcm_server_stats_free(const struct memcache_ctxt *ctxt, struct memcache_server_stats *s) {
  if (s->version != NULL)
    ctxt->mcFree(s->version);
  ctxt->mcFree(s);
}


static struct memcache_server_stats *
mcm_server_stats_new(const struct memcache_ctxt *ctxt) {
  struct memcache_server_stats *s;
  s = (struct memcache_server_stats *)ctxt->mcMalloc(sizeof(struct memcache_server_stats));
  if (s != NULL) {
    bzero(s, sizeof(struct memcache_server_stats));
  }

  return s;
}


static int
mcm_server_writable(struct memcache_ctxt *ctxt, struct memcache_server *ms, struct timeval *tv) {
#ifdef HAVE_SELECT
  int ret;

  FD_ZERO(&ms->fds);
  FD_SET(ms->fd, &ms->fds);

#ifdef DEBUG_MC_PROTO_ASSERT
  if (FD_ISSET(ms->fd, &ms->fds) == 0) {
    MCM_ERRX(MCM_ERR_ASSERT);
    return -1;
  }
#endif

#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "Begin select(2)'ing to see if fd %d is write(2)able.  Timeout %lu.%lu", ms->fd, (u_int64_t)tv->tv_sec, (u_int64_t)tv->tv_usec);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif

  /* Before we writev(2) anything, check to make sure the socket is
   * ready to accept data to be written out.  No sense in wastefully
   * calling writev(2) constantly in a loop. */
  ret = select(ms->fd + 1, NULL, &ms->fds, NULL, tv);
#ifdef DEBUG_MC_PROTO
  do {
    char *tm;
    size_t tml;
    tml = asprintf(&tm, "write(2)able fds: %d/%d", ret, FD_SETSIZE);
    if (tml > 0 && tm != NULL) {
      MCM_WARNX_MSG(MCM_ERR_TRACE, tm);
      free(tm);
    }
  } while (0);
#endif
  switch (ret) {
  case -1:
    switch (errno) {
    case EINTR:
      return 0;
    default:
      MCM_ERR(MCM_ERR_SYS_SELECT);
      break;
    }
  case 0:
    MCM_ERR_MSG_LVL(MCM_ERR_SYS_SELECT, "write select(2) call timed out", MCM_ERR_LVL_WARN);
    break;
  }

#ifdef DEBUG_MC_PROTO
  MCM_WARNX_MSG(MCM_ERR_TRACE, "End select(2)'ing to see if fd is write(2)able");
#endif

  return ret;
#else
  return 1;
#endif
}


int
mcm_set(struct memcache_ctxt *ctxt, struct memcache *mc,
	const char *key, const size_t key_len,
	const void *val, const size_t bytes,
	const time_t expire, const u_int16_t flags) {
  return mcm_storage_cmd(ctxt, mc, str_set_cmd, str_set_len, key, key_len, val, bytes, expire, flags);
}


struct memcache_server_stats *
mcm_stats(struct memcache_ctxt *ctxt, struct memcache *mc) {
  struct memcache_server *ms;
  struct memcache_server_stats *s, *ts;

  s = mcm_server_stats_new(ctxt);
  for (ms = mc->server_list.tqh_first; ms != NULL; ms = ms->entries.tqe_next) {
    ts = mcm_server_stats(ctxt, mc, ms);
    if (ts == NULL)
      continue;

    /* Merge the values from ts into s.  Any per-server specific data
     * is pulled from the last server. */
    s->pid = ts->pid;
    s->uptime = ts->uptime;
    s->time = ts->time;
    if (s->version == NULL && ts->version != NULL)
      s->version = mcm_strdup(ctxt, ts->version);

    s->rusage_user.tv_sec += ts->rusage_user.tv_sec;
    s->rusage_user.tv_usec += ts->rusage_user.tv_usec;
    if (s->rusage_user.tv_usec > 1000000) {
      s->rusage_user.tv_sec += s->rusage_user.tv_usec / 1000000;
      s->rusage_user.tv_usec -= 1000000 * (s->rusage_user.tv_usec / 1000000);
    }

    s->rusage_system.tv_sec += ts->rusage_system.tv_sec;
    s->rusage_system.tv_usec += ts->rusage_system.tv_usec;
    if (s->rusage_system.tv_usec > 1000000) {
      s->rusage_system.tv_sec += s->rusage_system.tv_usec / 1000000;
      s->rusage_system.tv_usec -= 1000000 * (s->rusage_system.tv_usec / 1000000);
    }

    s->curr_items += ts->curr_items;
    s->total_items += ts->total_items;
    s->bytes = s->bytes + ts->bytes;
    s->curr_connections += ts->curr_connections;
    s->total_connections += ts->total_connections;
    s->connection_structures += ts->connection_structures;
    s->cmd_get += ts->cmd_get;
#ifdef SEAN_HACKS
    s->cmd_refresh += ts->cmd_refresh;
#endif
    s->cmd_set += ts->cmd_set;
    s->get_hits += ts->get_hits;
    s->get_misses += ts->get_misses;
#ifdef SEAN_HACKS
    s->refresh_hits += ts->refresh_hits;
    s->refresh_misses += ts->refresh_misses;
#endif
    s->bytes_read += ts->bytes_read;
    s->bytes_written += ts->bytes_written;
    s->limit_maxbytes += ts->limit_maxbytes;

    mcm_server_stats_free(ctxt, ts);
  }

  return s;
}


static int
mcm_storage_cmd(struct memcache_ctxt *ctxt, struct memcache *mc,
		const char *cmd, const size_t cmd_len,
		const char *key, const size_t key_len,
		const void *val, const size_t bytes,
		const time_t expire, const u_int16_t flags) {
  struct memcache_server *ms;
  u_int32_t hash;
  size_t i;
  char *cp;
  struct iovec wv[11];

  /* Reset ctxt->errnum upon entry into memcache(3). */
  ctxt->errnum = 0;

  /* If there's only one server, don't bother actually hashing. */
  if (mc->num_live_servers == 1)
    hash = 0;
  else
    hash = mcm_hash_key(ctxt, key, key_len);

  ms = mcm_server_connect_next_avail(ctxt, mc, hash);
  if (ms == NULL)
    return -1;

  /* Reset the buffer so that I can chop it up to use it as a scratch
   * pad for converting numbers from binary to ASCII.  *hears the baby
   * jebus crying* */
  mcm_reset_buf(ms);
  ms->woff = 0;

  wv[0].iov_base = cmd;
  wv[0].iov_len = cmd_len;
  wv[1].iov_base = key;
  wv[1].iov_len = key_len;
  wv[2].iov_base = str_space;
  wv[2].iov_len = str_space_len;


  /* 5 == (2 ** 16).to_s.length */
  if (ms->size - ms->woff < 5) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 5));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return -2;
    } else {
      ms->size += MAX(ms->size, 5);
      ms->buf = cp;
    }
  }

  /* Convert flags to a string */
  i = (size_t)snprintf(&ms->buf[ms->woff], ms->size - ms->woff, "%u", flags);
  if (i < 1) {
    MCM_ERR(MCM_ERR_LIB_SNPRINTF);
    return -3;
  }

  wv[3].iov_base = &ms->buf[ms->woff];	/* Note where our flags string is located */
  wv[3].iov_len = i;
  wv[4].iov_base = wv[2].iov_base;	/* Add a space */
  wv[4].iov_len = wv[2].iov_len;

  ms->woff += ++i;;	/* advance cursor past trailing '\0' */


  /* 10 == (2 ** 32).to_s.length */
  if (ms->size - ms->woff < 10) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 10));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return -4;
    } else {
      ms->size += MAX(ms->size, 10);
      ms->buf = cp;
    }
  }

  /* Convert expiration time to a string */
  i = (size_t)snprintf(&ms->buf[ms->woff], ms->size - ms->woff, "%lu", (long int unsigned)expire);
  if (i < 1) {
    MCM_ERR(MCM_ERR_LIB_SNPRINTF);
    return -5;
  }

  wv[5].iov_base = &ms->buf[ms->woff];	/* Note where our flags string is located */
  wv[5].iov_len = i;
  wv[6].iov_base = wv[2].iov_base;	/* Add a space */
  wv[6].iov_len = wv[2].iov_len;

  ms->woff += ++i;	/* advance cursor past trailing '\0' */


  /* 10 == (2 ** 32).to_s.length */
  if (ms->size - ms->woff < 10) {
    cp = (char *)ctxt->mcRealloc(ms->buf, ms->size + MAX(ms->size, 10));
    if (cp == NULL) {
      MCM_ERR(MCM_ERR_MEM_REALLOC);
      return -6;
    } else {
      ms->size += MAX(ms->size, 10);
      ms->buf = cp;
    }
  }

  /* Convert bytes to a string */
  i = (size_t)snprintf(&ms->buf[ms->woff], ms->size - ms->woff, "%lu", (long int unsigned)bytes);
  if (i < 1) {
    MCM_ERR(MCM_ERR_LIB_SNPRINTF);
    return -7;
  }

  wv[7].iov_base = &ms->buf[ms->woff];	/* Note where our flags string is located */
  wv[7].iov_len = i;
  wv[8].iov_base = str_endl;	/* Newline */
  wv[8].iov_len = str_endl_len;

  ms->woff += ++i;	/* advance cursor past trailing '\0' */

  /* Add the data */
  wv[9].iov_base = val;
  wv[9].iov_len = bytes;

  /* Add another carriage return */
  wv[10].iov_base = str_endl;
  wv[10].iov_len = str_endl_len;

  if (mcm_server_send_cmd(ctxt, mc, ms, wv, 11) < 0)
    return -8;

  mcm_reset_buf(ms);
  cp = mcm_get_line(ctxt, mc, ms);
  if (cp != NULL && memcmp(cp, "STORED", MCM_CSTRLEN("STORED")) == 0) {
    /* Groovy Tuesday */
    return 0;
  } else if (cp != NULL && memcmp(cp, "NOT_STORED", MCM_CSTRLEN("NOT_STORED")) == 0) {
    /* Fuck beans.  That was them, wasn't it? */
    MCM_ERR_MSG(MCM_ERR_MC_STORE, cmd);
    return 1;
  }

  if (mc->num_live_servers == 0) {
    MCM_ERRX(MCM_ERR_MC_SERV_LIST);
    return 3;
  } else {
    MCM_ERRX(MCM_ERR_PROTO);
    return 2;
  }
}


char *
mcm_strdup(const struct memcache_ctxt *ctxt, const char *str) {
  return mcm_strndup(ctxt, str, strlen(str));
}


char *
mcm_strnchr(const struct memcache_ctxt *ctxt, mc_const char *str, int c, const size_t len) {
  char *cp;
  size_t i;

  for (cp = str, i = 0; i < len && *cp != '\0'; i++, cp++) {
    if (c == (int)*cp)
      return cp;
  }

  return '\0';
}


char *
mcm_strndup(const struct memcache_ctxt *ctxt, const char *str, const size_t len) {
  char *cp;

  cp = ctxt->mcMallocAtomic(len + 1);
  if (cp != NULL) {
    memcpy(cp, str, len);
    cp[len] = '\0';
  }

  return cp;
}


void
mcm_timeout(const struct memcache_ctxt *ctxt, struct memcache *mc, const int sec, const int msec) {
  mc->tv.tv_sec = sec;
  mc->tv.tv_usec = msec;
}


const u_int32_t
mcm_vernum(const struct memcache_ctxt *ctxt) {
  return MEMCACHE_VERNUM;
}


const char *
mcm_version(const struct memcache_ctxt *ctxt) {
  return MEMCACHE_VER;
}


/* BEGIN memcache memory API */
int
mcErrGet(mcErrFunc *errFunc) {
  if (errFunc != NULL)
    *errFunc = mcGlobalCtxt.mcErr;

  return 0;
}


int
mcErrSetup(mcErrFunc errFunc) {
  return mcErrSetupCtxt(&mcGlobalCtxt, errFunc);
}


int
mcErrSetupCtxt(struct memcache_ctxt *ctxt, mcErrFunc errFunc) {
  if (ctxt == NULL || errFunc == NULL)
    return 1;

  ctxt->mcErr = errFunc;

  return 0;
}


int
mcMemGet(mcFreeFunc *freeFunc, mcMallocFunc *mallocFunc, mcMallocFunc *mallocAtomicFunc,
	 mcReallocFunc *reallocFunc) {
  if (freeFunc != NULL)
    *freeFunc = mcGlobalCtxt.mcFree;

  if (mallocFunc != NULL)
    *mallocFunc = mcGlobalCtxt.mcMalloc;

  if (mallocAtomicFunc != NULL)
    *mallocAtomicFunc = mcGlobalCtxt.mcMallocAtomic;

  if (reallocFunc != NULL)
    *reallocFunc = mcGlobalCtxt.mcRealloc;

  return 0;
}


struct memcache_ctxt *
mcMemNewCtxt(mcFreeFunc freeFunc, mcMallocFunc mallocFunc, mcMallocFunc mallocAtomicFunc,
	     mcReallocFunc reallocFunc) {
  struct memcache_ctxt *ctxt;

  if (freeFunc == NULL || mallocFunc == NULL || reallocFunc == NULL)
    return NULL;

  ctxt = mallocFunc(sizeof(struct memcache_ctxt));
  if (ctxt != NULL) {
    bzero(ctxt, sizeof(struct memcache_ctxt));
    if (mcMemSetupCtxt(ctxt, freeFunc, mallocFunc, mallocAtomicFunc, reallocFunc) != 0) {
      bzero(ctxt, sizeof(struct memcache_ctxt));
      freeFunc(ctxt);
      return NULL;
    }

    /* Install our default error handler */
    ctxt->mcErr = mcm_err_func;
  }
  return ctxt;
}


void
mcMemFreeCtxt(struct memcache_ctxt *ctxt) {
  mcFreeFunc freeFunc;

  if (ctxt == NULL || ctxt->mcFree == NULL)
    return;

  freeFunc = ctxt->mcFree;
  freeFunc(ctxt);
}

int
mcMemSetup(mcFreeFunc freeFunc, mcMallocFunc mallocFunc,
	   mcMallocFunc mallocAtomicFunc, mcReallocFunc reallocFunc) {
  return mcMemSetupCtxt(&mcGlobalCtxt, freeFunc, mallocFunc, mallocAtomicFunc, reallocFunc);
}


int
mcMemSetupCtxt(struct memcache_ctxt *ctxt, mcFreeFunc freeFunc, mcMallocFunc mallocFunc,
	       mcMallocFunc mallocAtomicFunc, mcReallocFunc reallocFunc) {
  if (ctxt == NULL || freeFunc == NULL || mallocFunc == NULL || reallocFunc == NULL)
    return(1);

  ctxt->mcFree = freeFunc;
  ctxt->mcMalloc = mallocFunc;
  ctxt->mcMallocAtomic = (mallocAtomicFunc != NULL ? mallocAtomicFunc : mallocFunc);
  ctxt->mcRealloc = reallocFunc;

  return 0;
}
/* END memcache memory API */
